import redis
import kadabra
import json

from mock import MagicMock, mock

host = "host"
port = 1234
db = 0
logger = "testlogger"
queue_key = 'queue_key'
inprogress_key = 'inprogress_key'

def get_unit():
    return kadabra.channels.RedisChannel(host, port, db, logger, queue_key,
            inprogress_key)

@mock.patch('redis.StrictRedis')
def test_ctor(mock_redis):
    mock_redis.return_value = "test"

    channel = get_unit()

    mock_redis.assert_called_with(host=host, port=port, db=db)

    assert channel.client == "test"
    assert channel.logger.name == logger
    assert channel.queue_key == queue_key
    assert channel.inprogress_key == inprogress_key

def test_send():
    serialized = {"name": "value"}

    channel = get_unit()
    metrics = kadabra.Metrics([], [], [])

    metrics.serialize = MagicMock(return_value=serialized)
    channel.client.lpush = MagicMock()

    channel.send(metrics)

    metrics.serialize.assert_called_with()
    channel.client.lpush.assert_called_with(queue_key, json.dumps(serialized))

@mock.patch('kadabra.Metrics.deserialize')
def test_receive_metrics(mock_deserialize):
    deserialized = "test"
    mock_deserialize.return_value = deserialized
    raw_metrics = {"name": "value"}
    encoded = json.dumps(raw_metrics)

    channel = get_unit()
    channel.client.brpoplpush = MagicMock(return_value=encoded)

    metrics = channel.receive()

    channel.client.brpoplpush.assert_called_with(queue_key, inprogress_key,
            timeout=10)
    mock_deserialize.assert_called_with(raw_metrics)
    assert metrics == deserialized

def test_receive_nometrics():
    raw_metrics = None

    channel = get_unit()
    channel.client.brpoplpush = MagicMock(return_value=raw_metrics)

    metrics = channel.receive()

    channel.client.brpoplpush.assert_called_with(queue_key, inprogress_key,
            timeout=10)
    assert metrics == None

def test_complete_successful():
    serialized = {"name": "value"}

    channel = get_unit()
    metrics = kadabra.Metrics([], [], [])

    metrics.serialize = MagicMock(return_value=serialized)

    channel.client.lrem = MagicMock(return_value=1)

    channel.complete(metrics)

    metrics.serialize.assert_called_with()
    channel.client.lrem.assert_called_with(inprogress_key, 1,\
            json.dumps(serialized))

def test_complete_unsuccessful():
    serialized = {"name": "value"}

    channel = get_unit()
    metrics = kadabra.Metrics([], [], [])

    metrics.serialize = MagicMock(return_value=serialized)

    channel.client.lrem = MagicMock(return_value=0)

    channel.complete(metrics)

    metrics.serialize.assert_called_with()
    channel.client.lrem.assert_called_with(inprogress_key, 1,\
            json.dumps(serialized))

@mock.patch('kadabra.Metrics.deserialize')
def test_in_progress(mock_deserialize):
    query_limit = 3
    raw = [json.dumps({"nameOne": "valueOne"}),
           json.dumps({"nameTwo": "valueTwo"}),
           json.dumps({"nameThree": "valueThree"})]
    deserialized = ['one', 'two', 'three']
    mock_deserialize.side_effect = deserialized

    channel = get_unit()
    channel.client.lrange = MagicMock(return_value=raw)

    in_progress = channel.in_progress(query_limit)

    channel.client.lrange.assert_called_with(inprogress_key, 0,\
            query_limit - 1)

    for r in raw:
        mock_deserialize.assert_any_call(json.loads(r))
    assert in_progress == deserialized



import datetime, threading, json

from .channels import RedisChannel
from .config import DEFAULT_CONFIG
from .metrics import Dimension, Counter, Timer, Metrics
from .utils import get_now

class Client(object):
    """Main client API for Kadabra. In conjunction with the
    :class:`~kadabra.client.Collector`, allows you to collect metrics from your
    application and queue them for publishing via a channel.

    Typically you will use like so::

        client = Client()
        metrics = client.get_collector()
        ...
        metrics.add_count("myCount", 1.0)
        ...
        metrics.set_timer("myTimer", datetime.timedelta(seconds=5))
        ...
        metrics.add_count("myCount", 1.0)
        ...
        client.send(metrics.close())

    :type configuration: dict
    :param configuration: Dictionary of configuration to use in place of the
                          defaults.
    """
    def __init__(self, configuration=None):
        config = DEFAULT_CONFIG.copy()
        if configuration:
            config.update(configuration)

        default_dimensions = config.get("CLIENT_DEFAULT_DIMENSIONS", {})
        self.default_dimensions = default_dimensions

        channel_type = config["CLIENT_CHANNEL_TYPE"]
        custom_channel_args = config["CLIENT_CHANNEL_ARGS"]
        if channel_type == 'redis':
            channel_type = RedisChannel
        else:
            raise Exception("Unrecognized channel type: '%s'" % channel_type)

        channel_args = channel_type.DEFAULT_ARGS.copy()
        if custom_channel_args:
            for k,v in custom_channel_args.items():
                channel_args[k] = v

        self.channel = channel_type(**channel_args)

        self.timestamp_format = config["CLIENT_TIMESTAMP_FORMAT"]

    def get_collector(self):
        """Return a :class:`~kadabra.client.Collector` initialized with any
        dimensions as specified by this Client's default dimensions. The
        collector can be used to gather metrics from your application
        code.
        
        :rtype: ~kadabra.client.Collector
        :returns: A :class:`~kadabra.client.Collector` instance.
        """
        return Collector(self.timestamp_format, **self.default_dimensions)

    def send(self, metrics):
        """Send a :class:`Metrics` instance to this client's configured channel
        so that it can be received and published by the agent. Note that a
        Metrics instance can be retrieved from a collector by calling its
        :meth:`~kadabra.client.Collector.close` method.

        :type metrics: ~kadabra.Metrics
        :param metrics: The :class:`Metrics` instance to be published.
        """
        self.channel.send(metrics)

class Collector(object):
    """A class for collecting metrics. Once initialized, instances of this
    class collect metrics by aggregating counts and keeping track of dimensions
    and timers. Typically you won't instantiate this class directly, but rather
    retrieve an instance from the Client's :meth:`~Client.get_collector`
    method.

    :class:`Counter`\s are floating point values aggregated over the lifetime
    of this object, and published as a single value (per counter name).

    :class:`Timer`\s will be a floating point value along with a unit.

    A Collector instance can be used to collect metrics until it is closed by
    calling its :meth:`~kadabra.client.Collector.close` method. After close()
    has been called, this object can be safely published without the
    possibility of "losing" additional metrics between the time it is closed
    and the time it is published.

    Although Collector objects are thread-safe (meaning the same object can be
    used by multiple threads), note that any threads that attempt to use a
    Collector instance after it has been closed will throw an exception.

    :type timestamp_format: string
    :param timestamp_format: The format for timestamps when serializing into a
                             :class:`Metrics` instance.

    :type dimensions: dict
    :param dimensions: Any dimensions that this object should be initialized
                       with.
    """
    def __init__(self, timestamp_format, **dimensions):
        self.counters = {}
        self.timers = {}
        self.dimensions = dimensions if dimensions else {}
        self.timestamp_format = timestamp_format

        self.closed = False
        self.lock = threading.Lock()

    def set_dimension(self, name, value):
        """Set a dimension for this Collector object. If it already exists, it
        will be overwritten with the new value.

        :type name: string
        :param name: The name of the dimension to set.

        :type value: string
        :param value: The value of the dimension to be set.

        :raises CollectorClosedError: If this Collector object has
                                      already been closed.
        """
        self.lock.acquire()
        try:
            if self.closed:
                raise CollectorClosedError()

            self.dimensions[name] = value
        finally:
            self.lock.release()

    def add_count(self, name, value, timestamp=None, metadata=None,
            replace_timestamp=False):
        """Add a new counter to this Collector object, or add the value to an
        existing counter if it already exists.

        :type name: string
        :param name: The name of the counter.

        :type value: float
        :param value: The floating point value to either initialize a new
                      counter with, or add to an existing one.

        :type timestamp: ~datetime.datetime
        :param timestamp: The timestamp to use for when this count was
                          recorded. If unspecified, defaults to now (in UTC).

        :type metadata: dict
        :param metadata: Any metadata to include with this counter as a
                        dictionary of strings to strings. These will be
                        included as unindexed fields for this counter in
                        certain metrics databases. Note that if you specify
                        this for an existing counter, it will completely
                        overwrite the existing metadata. However if you do not
                        specify it, the previous metadata for the counter will
                        remain unchanged.

        :type replace_timestamp: bool
        :param replace_timestamp: Whether to replace the exisiting timestamp
                                  for a counter if it already exists. This can
                                  be set to True if you want to update the
                                  timestamp when you add to an existing
                                  counter.

        :raises CollectorClosedError: If this Collector object has
                                      already been closed.
        """
        self.lock.acquire()
        try:
            if self.closed:
                raise CollectorClosedError()

            if timestamp is None:
                timestamp = get_now()
            
            md = metadata if metadata else {}
            if name not in self.counters:
                self.counters[name] = {
                        "metadata": md,
                        "timestamp": timestamp,
                        "value": float(value)
                }
            else:
                self.counters[name]["value"] = self.counters[name]["value"] +\
                        float(value)
                if md:
                    self.counters[name]["metadata"] = md
                if replace_timestamp:
                    self.counters[name]["timestamp"] = timestamp
        finally:
            self.lock.release()

    def set_timer(self, name, value, unit, timestamp=None, metadata=None):
        """Set a timer value for this Collector object using a
        :class:`~datetime.timedelta`. If it already exists, it will be
        overwritten with the new value.

        :type name: string
        :param name: The name of the timer to set.

        :type value: ~datetime.timedelta
        :param value: The value to use for the timer.

        :type unit: ~kadabra.Unit
        :param unit: The unit to use for this timer. Common units are specified
                     in :class:`kadabra.Units`.

        :type timestamp: ~datetime.datetime
        :param timestamp: The timestamp to use for when this timer was
                          recorded. If unspecified, defaults to now (in UTC).

        :type metadata: dict
        :param metadata: Any metadata to include with this timer as a
                        dictionary of strings to strings. These will be
                        included as unindexed fields for this timer in
                        certain metrics databases. Note that if you specify
                        this for an existing timer, it will completely
                        overwrite the existing metadata. However if you do not
                        specify it, the previous metadata for the timer will
                        remain unchanged.

        :raises CollectorClosedError: If this Collector object has
                                      already been closed.
        """
        self.lock.acquire()
        try:
            if self.closed:
                raise CollectorClosedError()

            if timestamp is None:
                timestamp = get_now()
            md = metadata if metadata else {}

            if name not in self.timers:
                self.timers[name] = {
                        "unit": unit,
                        "metadata": md,
                        "timestamp": timestamp,
                        "value": value
                }
            else:
                self.timers[name]["unit"] = unit
                self.timers[name]["timestamp"] = timestamp
                self.timers[name]["value"] = value
                if md:
                    self.timers[name]["metadata"] = md
        finally:
            self.lock.release()

    def close(self):
        """Close this Collector object and return an equivalent
        :class:`Metrics` object. After this method is called, you can no longer
        set dimensions, set timers, or add counts to this object.

        :rtype: ~kadabra.Metrics
        :returns: A :class:`Metrics` instance from the Collector's dimensions,
                  counters, and timers.

        :raises CollectorClosedError: Raised if this Collector object
                                      has already been closed.
        """
        self.lock.acquire()
        try:
            if self.closed:
                raise CollectorClosedError()
            self.closed = True

            dimensions = [Dimension(n, v)\
                    for n,v in self.dimensions.items()]
            counters = [Counter(n, c["timestamp"], c["metadata"], c["value"])\
                    for n,c in self.counters.items()]
            timers = [Timer(n, t["timestamp"], t["metadata"], t["value"],\
                    t["unit"]) for n,t in self.timers.items()]
            return Metrics(dimensions, counters, timers, self.timestamp_format)
        finally:
            self.lock.release()

class CollectorClosedError(BaseException):
    """Raised if you try to add metrics to or close a
    :class:`~kadabra.client.Collector` object that has already been closed."""
    def __init__(self):
        super(CollectorClosedError, self).__init__(\
                "Collector object has been closed")




import kadabra
import datetime

def test_get_datetime_from_timestamp_string():
    now = datetime.datetime.utcnow()
    now = now.replace(microsecond=0)
    timestamp_format = "%Y-%m-%dT%H:%M:%SZ"
    assert kadabra.utils.get_datetime_from_timestamp_string(
            now.strftime(timestamp_format), timestamp_format) == now



import kadabra
import influxdb
import datetime

from mock import MagicMock, mock

def merge_dicts(*dict_args):
    '''
    Given any number of dicts, shallow copy and merge into a new dict,
    precedence goes to key value pairs in latter dicts.
    '''
    result = {}
    for dictionary in dict_args:
        result.update(dictionary)
    return result

@mock.patch('influxdb.InfluxDBClient')
def test_ctor(mock_influxdb):
    mocked = "test"
    host = "host"
    port = 1234
    database = "db"
    timeout = 3
    mock_influxdb.return_value = mocked

    publisher = kadabra.publishers.InfluxDBPublisher(host, port, database,
            timeout)

    mock_influxdb.assert_called_with(host=host, port=port, database=database,
            timeout=timeout)
    assert publisher.client == mocked

@mock.patch('influxdb.InfluxDBClient')
def test_publish_no_points(mock_influxdb):
    mocked = "test"
    host = "host"
    port = 1234
    database = "db"
    timeout = 3
    mocked = mock_influxdb.return_value
    mocked.write_points = MagicMock()

    timestamp_format = "%Y-%m-%dT%H:%M:%SZ"
    metrics = MagicMock()

    dimensionOne = MagicMock()
    dimensionOne.name = "dimensionOneName"
    dimensionOne.value = "dimensionOneValue"

    dimensionTwo = MagicMock()
    dimensionTwo.name = "dimensionTwoName"
    dimensionTwo.value = "dimensionTwoValue"

    metrics.dimensions = [dimensionOne, dimensionTwo]
    metrics.counters = []
    metrics.timers = []

    publisher = kadabra.publishers.InfluxDBPublisher(host, port, database,
            timeout)
    publisher.publish(metrics)

    publisher.client.write_points.assert_has_calls([])

@mock.patch('influxdb.InfluxDBClient')
def test_publish(mock_influxdb):
    mocked = "test"
    host = "host"
    port = 1234
    database = "db"
    timeout = 3
    mocked = mock_influxdb.return_value
    mocked.write_points = MagicMock()

    timestamp_format = "%Y-%m-%dT%H:%M:%SZ"
    metrics = MagicMock()
    metrics.timestamp_format = timestamp_format

    dimensionOne = MagicMock()
    dimensionOne.name = "dimensionOneName"
    dimensionOne.value = "dimensionOneValue"

    dimensionTwo = MagicMock()
    dimensionTwo.name = "dimensionTwoName"
    dimensionTwo.value = "dimensionTwoValue"

    timerOne = MagicMock()
    timerOne.name = "timerOne"
    timerOne.value = datetime.timedelta(seconds=10);
    timerOne.timestamp = datetime.datetime.utcnow()
    timerOne.metadata = {"timerOneMdName": "timerOneMdValue"}
    timerOne.unit = MagicMock()
    timerOne.unit.name = "timerOneUnit"
    timerOne.unit.seconds_offset = 10

    timerTwo = MagicMock()
    timerTwo.name = "timerTwo"
    timerTwo.value = datetime.timedelta(seconds=20)
    timerTwo.timestamp = datetime.datetime.utcnow() +\
            datetime.timedelta(seconds=5)
    timerTwo.metadata = {"timerTwoMdName": "timerTwoMdValue"}
    timerTwo.unit = MagicMock()
    timerTwo.unit.name = "timerTwoUnit"
    timerTwo.unit.seconds_offset = 20

    counterOne = MagicMock()
    counterOne.name = "counterOne"
    counterOne.value = 1.0;
    counterOne.timestamp = datetime.datetime.utcnow()
    counterOne.metadata = {"counterOneMdName": "counterOneMdValue"}

    counterTwo = MagicMock()
    counterTwo.name = "counterTwo"
    counterTwo.value = 2.0;
    counterTwo.timestamp = datetime.datetime.utcnow() +\
            datetime.timedelta(seconds=10)
    counterTwo.metadata = {"counterTwoMdName": "counterTwoMdValue"}

    metrics.dimensions = [dimensionOne, dimensionTwo]
    metrics.counters = [counterOne, counterTwo]
    metrics.timers = [timerOne, timerTwo]

    tags = {
        dimensionOne.name: dimensionOne.value,
        dimensionTwo.name: dimensionTwo.value
    }
    expected_points = [
        {
            "measurement": timerOne.name,
            "tags": tags,
            "time": datetime.datetime.strftime(timerOne.timestamp,
                timestamp_format),
            "fields": merge_dicts(timerOne.metadata, {\
                "value": kadabra.utils.timedelta_total_seconds(timerOne.value)\
                 * timerOne.unit.seconds_offset,
                "unit": timerOne.unit.name})
        },
        {
            "measurement": timerTwo.name,
            "tags": tags,
            "time": datetime.datetime.strftime(timerTwo.timestamp,
                timestamp_format),
            "fields": merge_dicts(timerTwo.metadata, {\
                "value": kadabra.utils.timedelta_total_seconds(timerTwo.value)\
                * timerTwo.unit.seconds_offset,
                "unit": timerTwo.unit.name})
        },
        {
            "measurement": counterOne.name,
            "tags": tags,
            "time": datetime.datetime.strftime(counterOne.timestamp,
                timestamp_format),
            "fields": merge_dicts(counterOne.metadata, {"value":
                counterOne.value})
        },
        {
            "measurement": counterTwo.name,
            "tags": tags,
            "time": datetime.datetime.strftime(counterTwo.timestamp,
                timestamp_format),
            "fields": merge_dicts(counterTwo.metadata, {"value":
                counterTwo.value})
        }
    ]

    publisher = kadabra.publishers.InfluxDBPublisher(host, port, database,
            timeout)
    publisher.publish(metrics)

    publisher.client.write_points.assert_called_with(expected_points)




import kadabra
import pytest

from mock import MagicMock, mock, call

@mock.patch('kadabra.agent.ReceiverThread')
def test_ctor(mock_receiver_thread):
    channel = MagicMock()
    publisher = MagicMock()
    logger = MagicMock()
    num_threads = 3

    receiver_threads = [MagicMock(), MagicMock(), MagicMock()]
    mock_receiver_thread.side_effect = receiver_threads

    receiver = kadabra.agent.Receiver(channel, publisher, logger, num_threads)

    assert receiver.channel == channel
    assert receiver.publisher == publisher
    assert receiver.logger == logger
    assert receiver.num_threads == num_threads

    mock_receiver_thread.assert_has_calls(
            [call(channel, publisher, logger) for x in receiver_threads])
    for i in range(len(receiver_threads)):
        expected_name = "KadabraReceiver-%s" % str(i)
        assert receiver.threads[i].name == expected_name

@mock.patch('kadabra.agent.ReceiverThread')
def test_start(mock_receiver_thread):
    channel = MagicMock()
    publisher = MagicMock()
    logger = MagicMock()
    num_threads = 3

    thread_one = MagicMock()
    thread_one.start = MagicMock()
    thread_one.name = "threadOne"
    thread_two = MagicMock()
    thread_two.start = MagicMock()
    thread_two.name = "threadTwo"
    thread_three = MagicMock()
    thread_three.start = MagicMock()
    thread_three.name = "threadThree"
    threads = [thread_one, thread_two, thread_three]
    mock_receiver_thread.side_effect = threads

    receiver = kadabra.agent.Receiver(channel, publisher, logger, num_threads)
    receiver.start()

    for thread in threads:
        thread.start.assert_called_with()

@mock.patch('kadabra.agent.ReceiverThread')
def test_stop(mock_receiver_thread):
    channel = MagicMock()
    publisher = MagicMock()
    logger = MagicMock()
    num_threads = 3

    thread_one = MagicMock()
    thread_one.stop = MagicMock()
    thread_one.name = "threadOne"
    thread_two = MagicMock()
    thread_two.stop = MagicMock()
    thread_two.name = "threadTwo"
    thread_three = MagicMock()
    thread_three.stop = MagicMock()
    thread_three.name = "threadThree"
    threads = [thread_one, thread_two, thread_three]
    mock_receiver_thread.side_effect = threads

    receiver = kadabra.agent.Receiver(channel, publisher, logger, num_threads)
    receiver.stop()

    for thread in threads:
        thread.stop.assert_called_with()



import kadabra
import datetime

from mock import MagicMock, mock, call

NOW = datetime.datetime.utcnow()
STRPTIME = NOW - datetime.timedelta(seconds=30)

def test_ctor():
    channel = MagicMock()
    publisher = MagicMock()
    logger = MagicMock()
    frequency_seconds = MagicMock()
    threshold_seconds = MagicMock()
    query_limit = MagicMock()
    num_threads = MagicMock()

    nanny = kadabra.agent.Nanny(channel, publisher, logger, frequency_seconds,
            threshold_seconds, query_limit, num_threads)

    assert nanny.channel == channel
    assert nanny.publisher == publisher
    assert nanny.logger == logger
    assert nanny.frequency_seconds == frequency_seconds
    assert nanny.threshold_seconds == threshold_seconds
    assert nanny.query_limit == query_limit
    assert nanny.num_threads == num_threads

@mock.patch('kadabra.agent.NannyThread')
@mock.patch('kadabra.agent.Timer')
def test_start(mock_timer, mock_nanny_thread):
    channel = MagicMock()
    publisher = MagicMock()
    logger = MagicMock()
    frequency_seconds = MagicMock()
    threshold_seconds = MagicMock()
    query_limit = MagicMock()
    num_threads = 3
    mock_timer.start = MagicMock()

    thread_one = MagicMock()
    thread_one.start = MagicMock()
    thread_two = MagicMock()
    thread_two.start = MagicMock()
    thread_three = MagicMock()
    thread_three.start = MagicMock()

    nanny_threads = [thread_one, thread_two, thread_three]
    mock_nanny_thread.side_effect = nanny_threads

    nanny = kadabra.agent.Nanny(channel, publisher, logger, frequency_seconds,
            threshold_seconds, query_limit, num_threads)
    nanny.start()

    mock_nanny_thread.assert_has_calls(
            [call(channel, publisher, nanny.queue, logger) for x in
                nanny_threads])
    for i in range(len(nanny_threads)):
        expected_name = "KadabraNannyThread-%s" % str(i)
        assert nanny.threads[i].name == expected_name
        nanny_threads[i].start.assert_called_with()

@mock.patch('kadabra.agent.get_datetime_from_timestamp_string',\
        return_value=STRPTIME)
@mock.patch('kadabra.agent.Timer')
def test_run_nanny(mock_timer, mock_get_datetime_from_timestamp_string):
    metrics_one = MagicMock()
    timestamp_format = "%Y-%m-%dT%H:%M:%SZ"
    metrics_one.serialized_at = STRPTIME.strftime(timestamp_format)
    metrics_one.timestamp_format = timestamp_format
    metrics = [metrics_one]

    channel = MagicMock()
    channel.in_progress = MagicMock(return_value=metrics)
    publisher = MagicMock()
    logger = MagicMock()
    frequency_seconds = MagicMock()
    threshold_seconds = 25
    query_limit = MagicMock()
    num_threads = MagicMock()
    mock_timer.start = MagicMock()

    timer = mock_timer.return_value

    nanny = kadabra.agent.Nanny(channel, publisher, logger, frequency_seconds,
            threshold_seconds, query_limit, num_threads)
    nanny.queue = MagicMock()
    nanny.queue.put = MagicMock()
    nanny._run_nanny()

    channel.in_progress.assert_called_with(query_limit)
    nanny.queue.put.assert_called_with(metrics_one)
    mock_timer.assert_called_with(frequency_seconds, nanny._run_nanny)
    assert timer.name == "KadabraNanny"
    timer.start.assert_called_with()

@mock.patch('kadabra.agent.Timer')
def test_run_nanny_missing_serialized_at(mock_timer):
    metrics_one = MagicMock()
    metrics_one.serialized_at = None
    metrics = [metrics_one]

    channel = MagicMock()
    channel.in_progress = MagicMock(return_value=metrics)
    publisher = MagicMock()
    logger = MagicMock()
    frequency_seconds = MagicMock()
    threshold_seconds = MagicMock()
    query_limit = MagicMock()
    num_threads = MagicMock()
    mock_timer.start = MagicMock()

    timer = mock_timer.return_value

    nanny = kadabra.agent.Nanny(channel, publisher, logger, frequency_seconds,
            threshold_seconds, query_limit, num_threads)
    nanny.queue = MagicMock()
    nanny.queue.put = MagicMock()
    nanny._run_nanny()

    channel.in_progress.assert_called_with(query_limit)
    nanny.queue.put.assert_called_with(metrics_one)
    mock_timer.assert_called_with(frequency_seconds, nanny._run_nanny)
    assert timer.name == "KadabraNanny"
    timer.start.assert_called_with()

@mock.patch('kadabra.agent.Timer')
def test_run_nanny_no_metrics(mock_timer):
    metrics = []

    channel = MagicMock()
    channel.in_progress = MagicMock(return_value=metrics)
    publisher = MagicMock()
    logger = MagicMock()
    frequency_seconds = MagicMock()
    threshold_seconds = MagicMock()
    query_limit = MagicMock()
    num_threads = MagicMock()
    mock_timer.start = MagicMock()

    timer = mock_timer.return_value

    nanny = kadabra.agent.Nanny(channel, publisher, logger, frequency_seconds,
            threshold_seconds, query_limit, num_threads)
    nanny.queue = MagicMock()
    nanny.queue.put = MagicMock()
    nanny._run_nanny()

    channel.in_progress.assert_called_with(query_limit)
    nanny.queue.put.assert_has_calls([])
    mock_timer.assert_called_with(frequency_seconds, nanny._run_nanny)
    assert timer.name == "KadabraNanny"
    timer.start.assert_called_with()

@mock.patch('kadabra.agent.Timer')
def test_run_nanny_exception(mock_timer):
    channel = MagicMock()
    channel.in_progress = MagicMock()
    channel.in_progress.side_effect = Exception()
    publisher = MagicMock()
    logger = MagicMock()
    frequency_seconds = MagicMock()
    threshold_seconds = MagicMock()
    query_limit = MagicMock()
    num_threads = MagicMock()
    mock_timer.start = MagicMock()

    timer = mock_timer.return_value

    nanny = kadabra.agent.Nanny(channel, publisher, logger, frequency_seconds,
            threshold_seconds, query_limit, num_threads)
    nanny.queue = MagicMock()
    nanny.queue.put = MagicMock()
    nanny._run_nanny()

    channel.in_progress.assert_called_with(query_limit)

    nanny.queue.put.assert_has_calls([])
    mock_timer.assert_called_with(frequency_seconds, nanny._run_nanny)
    assert timer.name == "KadabraNanny"
    timer.start.assert_called_with()

def test_stop_timer_none():
    channel = MagicMock()
    publisher = MagicMock()
    logger = MagicMock()
    frequency_seconds = MagicMock()
    threshold_seconds = MagicMock()
    query_limit = MagicMock()
    num_threads = MagicMock()

    thread_one = MagicMock()
    thread_one.stop = MagicMock()
    thread_two = MagicMock()
    thread_two.stop = MagicMock()
    thread_three = MagicMock()
    thread_three.stop = MagicMock()
    threads = [thread_one, thread_two, thread_three]

    nanny = kadabra.agent.Nanny(channel, publisher, logger, frequency_seconds,
            threshold_seconds, query_limit, num_threads)
    nanny.threads = threads
    nanny.stop()

    for thread in threads:
        thread.stop.assert_called_with()

def test_stop_timer_none():
    channel = MagicMock()
    publisher = MagicMock()
    logger = MagicMock()
    frequency_seconds = MagicMock()
    threshold_seconds = MagicMock()
    query_limit = MagicMock()
    num_threads = MagicMock()

    thread_one = MagicMock()
    thread_one.stop = MagicMock()
    thread_two = MagicMock()
    thread_two.stop = MagicMock()
    thread_three = MagicMock()
    thread_three.stop = MagicMock()
    threads = [thread_one, thread_two, thread_three]

    nanny = kadabra.agent.Nanny(channel, publisher, logger, frequency_seconds,
            threshold_seconds, query_limit, num_threads)
    nanny.threads = threads
    nanny.timer = MagicMock()
    nanny.timer.cancel = MagicMock()
    nanny.stop()

    nanny.timer.cancel.assert_called_with()
    for thread in threads:
        thread.stop.assert_called_with()



DEFAULT_CONFIG = {
    "CLIENT_DEFAULT_DIMENSIONS": {},
    "CLIENT_TIMESTAMP_FORMAT": "%Y-%m-%dT%H:%M:%S.%fZ",
    "CLIENT_CHANNEL_TYPE" : "redis",
    "CLIENT_CHANNEL_ARGS" : None,
    "AGENT_LOGGER_NAME": "kadabra.agent",
    "AGENT_CHANNEL_TYPE" : "redis",
    "AGENT_CHANNEL_ARGS" : None,
    "AGENT_PUBLISHER_TYPE" : "debug",
    "AGENT_PUBLISHER_ARGS" : None,
    "AGENT_RECEIVER_THREADS" : 3,
    "AGENT_NANNY_FREQUENCY_SECONDS" : 30.0,
    "AGENT_NANNY_THRESHOLD_SECONDS" : 60.0,
    "AGENT_NANNY_QUERY_LIMIT": 5000,
    "AGENT_NANNY_THREADS": 3
}



import kadabra, sys

if (sys.version_info > (3, 0)):
    from queue import Empty
else:
    from Queue import Empty

from mock import MagicMock

def test_ctor():
    channel = MagicMock()
    publisher = MagicMock()
    queue = MagicMock()
    logger = MagicMock()

    nanny_thread = kadabra.agent.NannyThread(channel, publisher, queue, logger)

    assert nanny_thread.channel == channel
    assert nanny_thread.publisher == publisher
    assert nanny_thread.queue == queue
    assert nanny_thread.logger == logger

def test_run_once_no_metrics():
    channel = MagicMock()
    publisher = MagicMock()
    queue = MagicMock()
    logger = MagicMock()

    queue.get = MagicMock(return_value=None)
    publisher.publish = MagicMock()
    channel.complete = MagicMock()

    nanny_thread = kadabra.agent.NannyThread(channel, publisher, queue, logger)
    nanny_thread._run_once()

    queue.get.assert_called_with(timeout=10)
    publisher.publish.assert_has_calls([])
    channel.complete.assert_has_calls([])

def test_run_once():
    channel = MagicMock()
    publisher = MagicMock()
    queue = MagicMock()
    logger = MagicMock()

    queue.get = MagicMock()
    metrics = queue.get.return_value
    metrics.serialize = MagicMock()
    publisher.publish = MagicMock()
    channel.complete = MagicMock()

    nanny_thread = kadabra.agent.NannyThread(channel, publisher, queue, logger)
    nanny_thread._run_once()

    queue.get.assert_called_with(timeout=10)
    publisher.publish.assert_called_with(metrics)
    channel.complete.assert_called_with(metrics)

def test_run_once_empty_queue():
    channel = MagicMock()
    publisher = MagicMock()
    queue = MagicMock()
    logger = MagicMock()

    queue.get = MagicMock()
    queue.get.side_effect = Empty()
    publisher.publish = MagicMock()
    channel.complete = MagicMock()

    nanny_thread = kadabra.agent.NannyThread(channel, publisher, queue, logger)
    nanny_thread._run_once()

    queue.get.assert_called_with(timeout=10)
    publisher.publish.assert_has_calls([])
    channel.complete.assert_has_calls([])

def test_run_once_exception():
    channel = MagicMock()
    publisher = MagicMock()
    queue = MagicMock()
    logger = MagicMock()

    queue.get = MagicMock()
    metrics = queue.get.return_value
    metrics.serialize = MagicMock()
    publisher.publish = MagicMock()
    publisher.publish.side_effect = Exception()
    channel.complete = MagicMock()

    nanny_thread = kadabra.agent.NannyThread(channel, publisher, queue, logger)
    nanny_thread._run_once()

    queue.get.assert_called_with(timeout=10)
    publisher.publish.assert_called_with(metrics)
    channel.complete.assert_has_calls([])

def test_stop():
    channel = MagicMock()
    publisher = MagicMock()
    queue = MagicMock()
    logger = MagicMock()

    nanny_thread = kadabra.agent.NannyThread(channel, publisher, queue, logger)
    nanny_thread.stopped = False
    nanny_thread.stop()

    assert nanny_thread.stopped == True

def test_check_stopped():
    channel = MagicMock()
    publisher = MagicMock()
    queue = MagicMock()
    logger = MagicMock()

    nanny_thread = kadabra.agent.NannyThread(channel, publisher, queue, logger)
    assert nanny_thread._check_stopped() == False
    nanny_thread.stopped = True
    assert nanny_thread._check_stopped() == True

def test_run():
    channel = MagicMock()
    publisher = MagicMock()
    queue = MagicMock()
    logger = MagicMock()

    nanny_thread = kadabra.agent.NannyThread(channel, publisher, queue, logger)
    nanny_thread._run_once = MagicMock()
    nanny_thread._check_stopped = MagicMock()
    nanny_thread._check_stopped.side_effect = [False, True]

    nanny_thread.run()

    assert nanny_thread._check_stopped.call_count == 2
    assert nanny_thread._run_once.call_count == 1



import kadabra
import logging

from mock import MagicMock, mock

@mock.patch('logging.getLogger')
def test_ctor(mock_get_logger):
    logger = "test"
    logger_name = "loggerName"
    mock_get_logger.return_value = logger

    publisher = kadabra.publishers.DebugPublisher(logger_name)
    mock_get_logger.assert_called_with(logger_name)
    assert publisher.logger == logger

@mock.patch('logging.getLogger')
def test_ctor(mock_get_logger):
    serialized = "test"
    logger_name = "loggerName"
    mocked_logger = mock_get_logger.return_value
    mocked_logger.info = MagicMock()
    mocked_metrics = MagicMock()
    mocked_metrics.serialize = MagicMock(return_value=serialized)

    publisher = kadabra.publishers.DebugPublisher(logger_name)
    publisher.publish(mocked_metrics)

    mocked_metrics.serialize.assert_called_with()
    mocked_logger.info.assert_called_with(serialized)



import kadabra
import pytest

from mock import MagicMock, mock, call

def test_ctor_unrecognized_channel():
    with pytest.raises(Exception):
        kadabra.Agent(configuration={"AGENT_CHANNEL_TYPE": "ehjgrhjehe"})

def test_ctor_unrecognized_publisher():
    with pytest.raises(Exception):
        kadabra.Agent(configuration={"AGENT_PUBLISHER_TYPE": "ehjgrhjehe"})

@mock.patch('kadabra.agent.DebugPublisher')
@mock.patch('kadabra.agent.RedisChannel')
@mock.patch('kadabra.agent.Receiver')
@mock.patch('kadabra.agent.Nanny')
def test_ctor_defaults(mock_nanny, mock_receiver, mock_redis_channel,
        mock_debug_publisher):
    channel = "testChannel"
    channel_default_args = {"channelArg1": "1", "channelArg2": 2}
    mock_redis_channel.return_value = channel
    mock_redis_channel.DEFAULT_ARGS = channel_default_args

    publisher = "testPublisher"
    publisher_default_args = {"publisherArg1": "5", "publisherArg2": 6}
    mock_debug_publisher.return_value = publisher
    mock_debug_publisher.DEFAULT_ARGS = publisher_default_args

    receiver = "testReceiver"
    mock_receiver.return_value = receiver

    nanny = "testNanny"
    mock_nanny.return_value = nanny

    expected_receiver_threads =\
            kadabra.config.DEFAULT_CONFIG["AGENT_RECEIVER_THREADS"]
    expected_nanny_frequency_seconds =\
            kadabra.config.DEFAULT_CONFIG["AGENT_NANNY_FREQUENCY_SECONDS"]
    expected_nanny_threshold_seconds =\
            kadabra.config.DEFAULT_CONFIG["AGENT_NANNY_THRESHOLD_SECONDS"]
    expected_nanny_query_limit =\
            kadabra.config.DEFAULT_CONFIG["AGENT_NANNY_QUERY_LIMIT"]
    expected_nanny_num_threads =\
            kadabra.config.DEFAULT_CONFIG["AGENT_NANNY_THREADS"]

    agent = kadabra.Agent()

    mock_redis_channel.assert_called_with(**channel_default_args)
    mock_debug_publisher.assert_called_with(**publisher_default_args)
    mock_receiver.assert_called_with(
            channel,
            publisher,
            agent.logger,
            expected_receiver_threads)
    mock_nanny.assert_called_with(
            channel,
            publisher,
            agent.logger,
            expected_nanny_frequency_seconds,
            expected_nanny_threshold_seconds,
            expected_nanny_query_limit,
            expected_nanny_num_threads)
    assert agent.receiver == receiver
    assert agent.nanny == nanny

@mock.patch('kadabra.agent.InfluxDBPublisher')
@mock.patch('kadabra.agent.RedisChannel')
@mock.patch('kadabra.agent.Receiver')
@mock.patch('kadabra.agent.Nanny')
def test_ctor_custom_args(mock_nanny, mock_receiver, mock_redis_channel,
        mock_influxdb_publisher):
    channel = "testChannel"
    channel_default_args = {"channelArg1": "1", "channelArg2": 2}
    channel_custom_args = {"channelArg2": 7}
    combined_channel_args = channel_default_args.copy()
    combined_channel_args.update(channel_custom_args)

    mock_redis_channel.return_value = channel
    mock_redis_channel.DEFAULT_ARGS = channel_default_args

    publisher = "testPublisher"
    publisher_default_args = {"publisherArg1": "5", "publisherArg2": 6}
    publisher_custom_args = {"publisherArg2": 9}
    combined_publisher_args = publisher_default_args.copy()
    combined_publisher_args.update(publisher_custom_args)

    mock_influxdb_publisher.return_value = publisher
    mock_influxdb_publisher.DEFAULT_ARGS = publisher_default_args

    receiver = "testReceiver"
    mock_receiver.return_value = receiver

    nanny = "testNanny"
    mock_nanny.return_value = nanny

    agent = kadabra.Agent(configuration={
        "AGENT_CHANNEL_ARGS": combined_channel_args,
        "AGENT_PUBLISHER_TYPE": 'influxdb',
        "AGENT_PUBLISHER_ARGS": combined_publisher_args
        })

    mock_redis_channel.assert_called_with(**combined_channel_args)
    mock_influxdb_publisher.assert_called_with(**combined_publisher_args)

@mock.patch('kadabra.agent.DebugPublisher')
@mock.patch('kadabra.agent.RedisChannel')
@mock.patch('kadabra.agent.Receiver')
@mock.patch('kadabra.agent.Nanny')
@mock.patch('kadabra.agent.time.sleep')
def test_start(mock_sleep, mock_nanny, mock_receiver, mock_redis_channel,
        mock_debug_publisher):
    nanny = mock_nanny.return_value
    nanny.start = MagicMock()
    receiver = mock_receiver.return_value
    receiver.start = MagicMock()

    mock_thread_one = MagicMock()
    mock_thread_one.join = MagicMock()
    mock_thread_two = MagicMock()
    mock_thread_two.join = MagicMock()
    mock_thread_three = MagicMock()
    mock_thread_three.join = MagicMock()
    receiver.threads = [mock_thread_one, mock_thread_two,
            mock_thread_three]

    channel_default_args = {"channelArg1": "1", "channelArg2": 2}
    publisher_default_args = {"publisherArg1": "5", "publisherArg2": 6}

    mock_redis_channel.DEFAULT_ARGS = channel_default_args
    mock_debug_publisher.DEFAULT_ARGS = publisher_default_args

    agent = kadabra.Agent()
    agent._check_stopped = MagicMock()
    agent._check_stopped.side_effect = [False, True]
    agent.start()

    agent.receiver.start.assert_called_with()
    agent.nanny.start.assert_called_with()
    assert agent._check_stopped.call_count == 2
    mock_sleep.assert_has_calls([call(10)])

@mock.patch('kadabra.agent.DebugPublisher')
@mock.patch('kadabra.agent.RedisChannel')
@mock.patch('kadabra.agent.Receiver')
@mock.patch('kadabra.agent.Nanny')
@mock.patch('kadabra.agent.time.sleep')
def test_start(mock_sleep, mock_nanny, mock_receiver, mock_redis_channel,
        mock_debug_publisher):
    mock_sleep.side_effect = KeyboardInterrupt()

    nanny = mock_nanny.return_value
    nanny.start = MagicMock()
    receiver = mock_receiver.return_value
    receiver.start = MagicMock()

    mock_thread_one = MagicMock()
    mock_thread_one.join = MagicMock()
    mock_thread_two = MagicMock()
    mock_thread_two.join = MagicMock()
    mock_thread_three = MagicMock()
    mock_thread_three.join = MagicMock()
    receiver.threads = [mock_thread_one, mock_thread_two,
            mock_thread_three]

    channel_default_args = {"channelArg1": "1", "channelArg2": 2}
    publisher_default_args = {"publisherArg1": "5", "publisherArg2": 6}

    mock_redis_channel.DEFAULT_ARGS = channel_default_args
    mock_debug_publisher.DEFAULT_ARGS = publisher_default_args

    agent = kadabra.Agent()
    agent._check_stopped = MagicMock()
    agent._check_stopped.return_value = False
    agent.stop = MagicMock()
    agent.start()

    agent.receiver.start.assert_called_with()
    agent.nanny.start.assert_called_with()
    assert agent._check_stopped.call_count == 1
    agent.stop.assert_called_with()

@mock.patch('kadabra.agent.DebugPublisher')
@mock.patch('kadabra.agent.RedisChannel')
@mock.patch('kadabra.agent.Receiver')
@mock.patch('kadabra.agent.Nanny')
@mock.patch('kadabra.agent.time.sleep')
def test_check_stopped(mock_sleep, mock_nanny, mock_receiver, mock_redis_channel,
        mock_debug_publisher):
    mock_sleep.side_effect = KeyboardInterrupt()

    nanny = mock_nanny.return_value
    receiver = mock_receiver.return_value

    mock_thread_one = MagicMock()
    mock_thread_one.join = MagicMock()
    mock_thread_two = MagicMock()
    mock_thread_two.join = MagicMock()
    mock_thread_three = MagicMock()
    mock_thread_three.join = MagicMock()
    receiver.threads = [mock_thread_one, mock_thread_two,
            mock_thread_three]

    channel_default_args = {"channelArg1": "1", "channelArg2": 2}
    publisher_default_args = {"publisherArg1": "5", "publisherArg2": 6}

    mock_redis_channel.DEFAULT_ARGS = channel_default_args
    mock_debug_publisher.DEFAULT_ARGS = publisher_default_args

    agent = kadabra.Agent()
    agent._check_stopped() == False

@mock.patch('kadabra.agent.DebugPublisher')
@mock.patch('kadabra.agent.RedisChannel')
@mock.patch('kadabra.agent.Receiver')
@mock.patch('kadabra.agent.Nanny')
@mock.patch('kadabra.agent.time.sleep')
def test_stop(mock_sleep, mock_nanny, mock_receiver, mock_redis_channel,
        mock_debug_publisher):
    mock_sleep.side_effect = KeyboardInterrupt()

    nanny = mock_nanny.return_value
    nanny.stop = MagicMock()
    receiver = mock_receiver.return_value
    receiver.stop = MagicMock()

    mock_thread_one = MagicMock()
    mock_thread_one.join = MagicMock()
    mock_thread_two = MagicMock()
    mock_thread_two.join = MagicMock()
    mock_thread_three = MagicMock()
    mock_thread_three.join = MagicMock()
    receiver.threads = [mock_thread_one, mock_thread_two,
            mock_thread_three]

    channel_default_args = {"channelArg1": "1", "channelArg2": 2}
    publisher_default_args = {"publisherArg1": "5", "publisherArg2": 6}

    mock_redis_channel.DEFAULT_ARGS = channel_default_args
    mock_debug_publisher.DEFAULT_ARGS = publisher_default_args

    agent = kadabra.Agent()
    assert agent._check_stopped() == False
    agent.stop()
    assert agent._check_stopped() == True
    nanny.stop.assert_called_with()
    receiver.stop.assert_called_with()



import threading, logging, datetime, json, time, sys

if (sys.version_info > (3, 0)):
    from queue import Queue, Empty
else:
    from Queue import Queue, Empty

from threading import Timer

from .channels import RedisChannel
from .publishers import DebugPublisher, InfluxDBPublisher
from .utils import get_now, get_datetime_from_timestamp_string,\
                   timedelta_total_seconds

from .config import DEFAULT_CONFIG

class Agent(object):
    """Reads metrics from a channel and publishes them (see
    :ref:`api-publishers`). The agent will spin up threads which listen
    to the configured channel and attempt to publish the metrics using the
    configured publisher. The agent will also periodically monitor the metrics
    that have been in progress for a while and attempt to republish them.
    Because the agent is meant to run indefinitely, side by side with your
    application, it should be configured and started in a separate, dedicated
    process.

    Internally this object just manages a :class:`~kadabra.agent.Receiver` and
    :class:`~kadabra.agent.Nanny`.

    :type configuration: dict
    :param configuration: Dictionary of configuration to use in place of the
                          defaults.
    """
    def __init__(self, configuration=None):
        config = DEFAULT_CONFIG.copy()
        if configuration:
            config.update(configuration)

        self.logger = logging.getLogger(config["AGENT_LOGGER_NAME"])
        self.logger.info("Initializing Agent with config: %s" % str(config))

        channel_type = config["AGENT_CHANNEL_TYPE"]
        custom_channel_args = config["AGENT_CHANNEL_ARGS"]
        if channel_type == 'redis':
            channel_type = RedisChannel
        else:
            raise Exception("Unrecognized channel type: '%s'" % channel_type)

        channel_args = channel_type.DEFAULT_ARGS.copy()
        if custom_channel_args:
            for k,v in custom_channel_args.items():
                channel_args[k] = v
        channel = channel_type(**channel_args)

        publisher_type = config["AGENT_PUBLISHER_TYPE"]
        custom_publisher_args = config["AGENT_PUBLISHER_ARGS"]
        if publisher_type == 'debug':
            publisher_type = DebugPublisher
        elif publisher_type == 'influxdb':
            publisher_type = InfluxDBPublisher
        else:
            raise Exception("Unrecognized publisher type: '%s'" %\
                    publisher_type)

        publisher_args = publisher_type.DEFAULT_ARGS.copy()
        if custom_publisher_args:
            for k,v in custom_publisher_args.items():
                publisher_args[k] = v
        publisher = publisher_type(**publisher_args)

        receiver_threads = config["AGENT_RECEIVER_THREADS"]

        nanny_frequency_seconds = config["AGENT_NANNY_FREQUENCY_SECONDS"]
        nanny_threshold_seconds = config["AGENT_NANNY_THRESHOLD_SECONDS"]
        nanny_query_limit = config["AGENT_NANNY_QUERY_LIMIT"]
        nanny_num_threads = config["AGENT_NANNY_THREADS"]

        self.receiver = Receiver(channel, publisher, self.logger,
                receiver_threads)
        self.nanny = Nanny(channel, publisher, self.logger,
                nanny_frequency_seconds, nanny_threshold_seconds,
                nanny_query_limit, nanny_num_threads)

        self.stopped = False

    def start(self):
        """Start the agent. It will receive metrics from the channel, publish
        them, and attempt to republish metrics that have been pending for a
        long time (in the case of publishing failures). The agent runs until
        stopped; thus, you should call this method from a dedicated Python
        process, as it will block until the process is killed, a keyboard
        interrupt is detected, or the :meth:`~kadabra.Agent.stop` method is
        called."""
        self.logger.info("Starting agent...")
        self.receiver.start()
        self.nanny.start()

        try:
            while not self._check_stopped():
                time.sleep(10)
        except KeyboardInterrupt:
            self.stop()

    def stop(self, *args, **kwargs):
        """Stop the Agent gracefully, ensuring that any pending publish
        attempts are finished. This method accepts arbitrary arguments so that
        it can be called from any context (such as a signal handler)."""
        self.logger.info("Gracefully stopping the agent... "
                         "This may take up to 10 seconds.")
        self.stopped = True
        self.nanny.stop()
        self.receiver.stop()

    def _check_stopped(self):
        """Determines if the agent has been stopped. This is used internally to
        run the Agent continuously until stopped.

        :rtype: bool
        :returns: True if the Agent has been stopped, False otherwise.
        """
        return self.stopped

class Receiver(object):
    """Manages :class:`~kadabra.agent.ReceiverThread`\s which receive metrics
    from the channel, move them from the queue to in-progress, and attempt to
    publish them. Publishing failures will result in the metrics remaining
    in-progress and getting picked up by the :class:`~kadabra.agent.Nanny`
    which will attempt to republish them.

    :type channel: :ref:`api-channels`
    :param channel: The channel to read metrics from. See
                    :ref:`api-channels`.

    :type publisher: :ref:`api-publishers`
    :param publisher: The publisher to use for publishing metrics. See
                      :ref:`api-publishers`.

    :type logger: ~logging.Logger
    :param logger: The logger to use.

    :type num_threads: integer
    :param num_threads: The number of threads to use for publishing metrics.
    """
    def __init__(self, channel, publisher, logger, num_threads):
        self.channel = channel
        self.publisher = publisher
        self.logger = logger
        self.num_threads = num_threads

        self.threads = []
        for i in range(self.num_threads):
            name = "KadabraReceiver-%s" % str(i)
            receiver_thread = ReceiverThread(self.channel, self.publisher,\
                    self.logger)
            receiver_thread.name = name
            self.threads.append(receiver_thread)

    def start(self):
        """Start the receiver by starting up each
        :class:`~kadabra.agent.ReceiverThread`."""
        for thread in self.threads:
            self.logger.info("Starting %s" % thread.name)
            thread.start()

    def stop(self):
        """Stop the receiver by stopping each
        :class:`~kadabra.agent.ReceiverThread`."""
        self.logger.info("Stopping Receiver...")
        for thread in self.threads:
            thread.stop()

class ReceiverThread(threading.Thread):
    """Listens to a channel for metrics and publishes them.

    :type channel: :ref:`api-channels`
    :param channel: The channel to read metrics from. See
                    :ref:`api-channels`.

    :type publisher: :ref:`api-publishers`
    :param publisher: The publisher to use for publishing metrics. See
                      :ref:`api-publishers`.

    :type logger: ~logging.Logger
    :param logger: The logger to use.
    """
    def __init__(self, channel, publisher, logger):
        super(ReceiverThread, self).__init__()
        self.channel = channel
        self.publisher = publisher
        self.logger = logger

        self.stopped = False

    def stop(self):
        """Stops this this thread, ensuring that the current run will be the
        last one."""
        self.logger.info("Stopping %s..." % self.name)
        self.stopped = True

    def run(self):
        """Run this thread until stopped."""
        while not self._check_stopped():
            self._run_once()
        self.logger.info("Stopped %s." % self.name)

    def _run_once(self):
        """Runs this thread once. It will receive a message from the channel
        containing the metrics, publish them using the publisher, and mark the
        metrics as complete in the channel."""
        try:
            metrics = self.channel.receive() # This could be blocking
            if metrics is not None:
                self.logger.debug("Publishing metrics: %s" %
                        metrics.serialize())
                self.publisher.publish(metrics)
                self.channel.complete(metrics)
        except:
            self.logger.warn("Receiver thread encountered exception",\
                    exc_info=1)

    def _check_stopped(self):
        """Determines if this thread has been stopped. This is used internally
        to run the thread continuously until stopped.

        :rtype: bool
        :returns: True if the thread has been stopped, False otherwise.
        """
        return self.stopped

class Nanny(object):
    """Monitors metrics that have been in-progress for a long time and attemps
    to republish them. This object will periodically query objects in the
    in-progress queue, and try to republish them if the time between now and
    when they were serialized is greater than a threshold (indicating the first
    attempt to publish the metrics failed). It will grab the first X elements
    from the in-progress queue (where X is a configured value) and add them to
    a queue, which :class:`~kadabra.agent.NannyThread`\s will read from and
    attempt to republish. If metrics are successfully published they will be
    marked as complete.

    :type channel: :ref:`api-channels`
    :param channel: The channel to monitor.

    :type publisher: :ref:`api-publishers`
    :param publisher: The publisher to use for republishing metrics.

    :type logger: ~logging.Logger
    :param logger: The logger to use.

    :type frequency_seconds: integer
    :param frequency_seconds: How often the Nanny should query the in_progress
                              queue.

    :type threshold_seconds: integer
    :param threshold_seconds: The threshold seconds to determine if metrics
                              should be attempted to be republished.

    :type query_limit: integer
    :param query_limit: The maximum number of elements to query from the
                        in-progress queue for any given Nanny run. This is
                        necessary because the in-progress queue will constantly
                        be changing. Thus nanny needs to take a "snapshot"
                        rather than iterate through the queue.

    :type num_threads: integer
    :param num_threads: The number of :class:`~kadabra.agent.NannyThread`\s to
                        use for republishing.
    """
    def __init__(self, channel, publisher, logger, frequency_seconds,
            threshold_seconds, query_limit, num_threads):
        self.channel = channel
        self.publisher = publisher
        self.logger = logger
        self.frequency_seconds = frequency_seconds
        self.threshold_seconds = threshold_seconds
        self.query_limit = query_limit
        self.num_threads = num_threads

        self.queue = Queue()
        self.threads = []
        self.timer = None

    def start(self):
        """Start the nanny by starting up each
        :class:`~kadabra.agent.NannyThread`."""
        self.threads = []
        for i in range(self.num_threads):
            name = "KadabraNannyThread-%s" % str(i)
            nanny_thread = NannyThread(self.channel, self.publisher,
                    self.queue, self.logger)
            nanny_thread.name = name
            self.threads.append(nanny_thread)
            nanny_thread.start()

        timer = Timer(self.frequency_seconds, self._run_nanny)
        timer.name = "KadabraNanny"
        self.timer = timer
        timer.start()

    def stop(self):
        """Stop the nanny by stopping the Nanny from listening to the channela
        nd by stopping each :class:`~kadabra.agent.NannyThread`."""
        self.logger.info("Stopping Nanny...")
        if self.timer is not None:
            self.timer.cancel()
        for thread in self.threads:
            thread.stop()

    def _run_nanny(self):
        """Runs the nanny. It will check the channel's in-progress queue at the
        configured frequency, and add any in-progress items to an internal
        queue, which the :class:`~kadabra.agent.NannyThread`\s will listen to
        and attempt to republish metrics from."""
        try:
            self.logger.debug("Running nanny")
            in_progress = self.channel.in_progress(self.query_limit)

            if len(in_progress) == 0:
                self.logger.debug("No metrics found in progress.")

            for metrics in in_progress:
                should_republish = False
                if metrics.serialized_at is not None:
                    now = get_now()
                    serialized_at = get_datetime_from_timestamp_string(
                            metrics.serialized_at, metrics.timestamp_format)
                    delta = timedelta_total_seconds(now - serialized_at)
                    self.logger.debug(\
                            "serialized_at: %s, now: %s, delta: %s" %\
                        (str(metrics.serialized_at), str(now), delta))
                    should_republish = delta > self.threshold_seconds
                else:
                    self.logger.warn("Metrics is missing serialized_at, "
                            "something is wrong with the channel. "
                            "Attempting to republish anyway")
                    should_republish = True
                if should_republish:
                    self.queue.put(metrics)
        except:
            self.logger.warn("Encountered exception trying to get "
                    "in-progress metrics", exc_info=1)
        finally:
            timer = Timer(self.frequency_seconds, self._run_nanny)
            timer.name = "KadabraNanny"
            self.timer = timer
            timer.start()

class NannyThread(threading.Thread):
    """Listens to a queue for metrics that have been in progress for a long
    time and attempts to republish them. If the publishing is successful,
    marks the metrics as complete.

    :type channel: :ref:`api-channels`
    :param channel: The channel to mark the metrics as complete upon successful
                    publishing.

    :type publisher: :ref:`api-publishers`
    :param publisher: The publisher to be used to publish the metrics object.

    :type queue: Queue
    :param queue: The queue to monitor for metrics to republish.

    :type logger: ~logging.Logger
    :param logger: The :class:`Logger` to log messages to.
    """
    def __init__(self, channel, publisher, queue, logger):
        super(NannyThread, self).__init__()
        self.channel = channel
        self.publisher = publisher
        self.queue = queue
        self.logger = logger

        self.stopped = False

    def stop(self):
        """Stops this this thread, ensuring that the current run will be the
        last one."""
        self.logger.info("Stopping %s..." % self.name)
        self.stopped = True

    def run(self):
        """Run this thread until stopped."""
        while not self._check_stopped():
            self._run_once()
        self.logger.info("Stopped %s." % self.name)

    def _run_once(self):
        """Listen to the queue for metrics to publish and attempt to publish
        them and mark them as complete."""
        try:
            metrics = self.queue.get(timeout=10)
            if metrics is not None:
                self.logger.debug("Publishing metrics: %s" %\
                        metrics.serialize())
                self.publisher.publish(metrics)
                self.channel.complete(metrics)
        except Empty:
            pass
        except:
            self.logger.warn("Nanny thread encountered exception",\
                    exc_info=1)

    def _check_stopped(self):
        """Determines if this thread has been stopped. This is used internally
        to run the thread continuously until stopped.

        :rtype: bool
        :returns: True if the thread has been stopped, False otherwise.
        """
        return self.stopped



import kadabra
import pytest
import datetime

from mock import MagicMock, mock, call

NOW = datetime.datetime.utcnow()

def test_client_ctor_unrecognized_channel():
    with pytest.raises(Exception):
        kadabra.Client(configuration={"CLIENT_CHANNEL_TYPE": "grenjiweroni"})

# http://www.voidspace.org.uk/python/mock/patch.html#where-to-patch
@mock.patch('kadabra.client.RedisChannel')
def test_client_ctor_defaults(mock_redis_channel):
    channel = "test"
    channel_default_args = {"arg1": "1", "arg2": 2}

    mock_redis_channel.return_value = channel
    mock_redis_channel.DEFAULT_ARGS = channel_default_args

    client = kadabra.Client()

    mock_redis_channel.assert_called_with(**channel_default_args)
    assert client.default_dimensions == {}
    assert client.channel == channel

@mock.patch('kadabra.client.RedisChannel')
def test_client_ctor_custom_channel_args(mock_redis_channel):
    channel = "test"
    channel_default_args = {"arg1": "1", "arg2": 2}
    channel_custom_args = {"arg2": 3}
    combined_args = channel_default_args.copy()
    combined_args.update(channel_custom_args)

    mock_redis_channel.return_value = channel
    mock_redis_channel.DEFAULT_ARGS = channel_default_args

    client = kadabra.Client(\
            configuration={"CLIENT_CHANNEL_ARGS": combined_args})

    mock_redis_channel.assert_called_with(**combined_args)
    assert client.channel == channel

@mock.patch('kadabra.client.Collector')
@mock.patch('kadabra.client.RedisChannel')
def test_client_get_collector_no_default_dimensions(mock_redis_channel,
        mock_collector):
    timestamp_format = "timestamp_format"
    expected_collector = "test_collector"
    channel = "test"
    channel_default_args = {"arg1": "1", "arg2": 2}

    mock_redis_channel.return_value = channel
    mock_redis_channel.DEFAULT_ARGS = channel_default_args

    mock_collector.return_value = expected_collector

    client = kadabra.Client(\
            configuration={"CLIENT_TIMESTAMP_FORMAT": timestamp_format})
    collector = client.get_collector()

    mock_collector.assert_called_with(timestamp_format)
    assert collector == expected_collector

@mock.patch('kadabra.client.Collector')
@mock.patch('kadabra.client.RedisChannel')
def test_client_get_collector_default_dimensions(mock_redis_channel,
        mock_collector):
    timestamp_format = "timestamp_format"
    default_dimensions = {"one": "o", "two": "t"}
    expected_collector = "test_collector"
    channel = "test"
    channel_default_args = {"arg1": "1", "arg2": 2}

    mock_redis_channel.return_value = channel
    mock_redis_channel.DEFAULT_ARGS = channel_default_args

    mock_collector.return_value = expected_collector

    client = kadabra.Client(\
            configuration={"CLIENT_DEFAULT_DIMENSIONS": default_dimensions,
                "CLIENT_TIMESTAMP_FORMAT": timestamp_format})
    collector = client.get_collector()

    mock_collector.assert_called_with(timestamp_format, **default_dimensions)
    assert collector == expected_collector

@mock.patch('kadabra.client.RedisChannel')
def test_client_send(mock_redis_channel):
    to_send = "to_send"
    channel_default_args = {"arg1": "1", "arg2": 2}

    mock_redis_channel.DEFAULT_ARGS = channel_default_args
    mock_redis_channel.send = MagicMock()
    mock_channel_instance = mock_redis_channel.return_value

    client = kadabra.Client()
    client.send(to_send)

    mock_channel_instance.send.assert_called_with(to_send)

def test_collector_ctor_no_dimensions():
    timestamp_format = "timestamp_format"
    collector = kadabra.client.Collector(timestamp_format)

    assert collector.counters == {}
    assert collector.timers == {}
    assert collector.dimensions == {}
    assert collector.closed == False
    assert hasattr(collector, "lock")

def test_collector_ctor_dimensions():
    timestamp_format = "timestamp_format"
    dimensions = {"dimensionOne": "d1", "dimensionTwo": "d2"}
    collector = kadabra.client.Collector(timestamp_format, **dimensions)

    assert collector.counters == {}
    assert collector.timers == {}
    assert collector.dimensions == dimensions
    assert collector.closed == False
    assert hasattr(collector, "lock")

class MockLock(object):
    def __init__(self):
        self.acquire = MagicMock(return_value=True)
        self.release = MagicMock(return_value=True)

def test_collector_close_closed():
    timestamp_format = "timestamp_format"
    collector = kadabra.client.Collector(timestamp_format)
    collector.lock = MockLock()
    collector.close()
    with pytest.raises(kadabra.client.CollectorClosedError):
        collector.close()
    assert len(collector.lock.acquire.mock_calls) == 2
    assert len(collector.lock.release.mock_calls) == 2

def test_collector_set_dimension_closed():
    timestamp_format = "timestamp_format"
    collector = kadabra.client.Collector(timestamp_format)
    collector.lock = MockLock()
    collector.close()
    with pytest.raises(kadabra.client.CollectorClosedError):
        collector.set_dimension("name", "value")
    assert len(collector.lock.acquire.mock_calls) == 2
    assert len(collector.lock.release.mock_calls) == 2

def test_collector_add_count_closed():
    timestamp_format = "timestamp_format"
    collector = kadabra.client.Collector(timestamp_format)
    collector.lock = MockLock()
    collector.close()
    with pytest.raises(kadabra.client.CollectorClosedError):
        collector.add_count("name", 1.0)
    assert len(collector.lock.acquire.mock_calls) == 2
    assert len(collector.lock.release.mock_calls) == 2

def test_collector_set_timer_closed():
    timestamp_format = "timestamp_format"
    collector = kadabra.client.Collector(timestamp_format)
    collector.lock = MockLock()
    collector.close()
    with pytest.raises(kadabra.client.CollectorClosedError):
        collector.set_timer("name", datetime.timedelta(seconds=5),
                kadabra.Units.MILLISECONDS)
    assert len(collector.lock.acquire.mock_calls) == 2
    assert len(collector.lock.release.mock_calls) == 2

def test_collector_set_dimension():
    dimension_name = "name"
    dimension_value = "value"
    timestamp_format = "timestamp_format"
    collector = kadabra.client.Collector(timestamp_format)
    collector.lock = MockLock()

    collector.set_dimension(dimension_name, dimension_value)

    assert len(collector.lock.acquire.mock_calls) == 1
    assert len(collector.lock.release.mock_calls) == 1
    assert collector.dimensions[dimension_name] == dimension_value

def test_collector_set_dimension_override():
    dimension_name = "name"
    dimension_value = "newValue"
    timestamp_format = "timestamp_format"
    collector = kadabra.client.Collector(timestamp_format, name="oldValue")
    collector.lock = MockLock()

    collector.set_dimension(dimension_name, dimension_value)

    assert len(collector.lock.acquire.mock_calls) == 1
    assert len(collector.lock.release.mock_calls) == 1
    assert collector.dimensions[dimension_name] == dimension_value

@mock.patch('kadabra.client.get_now', return_value=NOW)
def test_collector_add_count(mock_get_now):
    count_name = "name"
    count_value = 1.0
    timestamp_format = "timestamp_format"

    collector = kadabra.client.Collector(timestamp_format)
    collector.lock = MockLock()

    collector.add_count(count_name, count_value)

    assert len(collector.lock.acquire.mock_calls) == 1
    assert len(collector.lock.release.mock_calls) == 1
    assert collector.counters[count_name]["value"] == count_value
    assert collector.counters[count_name]["metadata"] == {}
    assert collector.counters[count_name]["timestamp"] == NOW

@mock.patch('kadabra.client.get_now', return_value=NOW)
def test_collector_add_count_timestamp(mock_get_now):
    count_name = "name"
    count_value = 1.0
    timestamp_format = "timestamp_format"
    timestamp = NOW + datetime.timedelta(seconds=30)

    collector = kadabra.client.Collector(timestamp_format)
    collector.lock = MockLock()

    collector.add_count(count_name, count_value, timestamp=timestamp)

    assert len(collector.lock.acquire.mock_calls) == 1
    assert len(collector.lock.release.mock_calls) == 1
    assert collector.counters[count_name]["value"] == count_value
    assert collector.counters[count_name]["metadata"] == {}
    assert collector.counters[count_name]["timestamp"] == timestamp

@mock.patch('kadabra.client.get_now', return_value=NOW)
def test_collector_add_count_metadata(mock_get_now):
    count_name = "name"
    count_value = 1.0
    count_metadata = {"name": "value"}
    timestamp_format = "timestamp_format"

    collector = kadabra.client.Collector(timestamp_format)
    collector.lock = MockLock()

    collector.add_count(count_name, count_value, metadata=count_metadata)

    assert len(collector.lock.acquire.mock_calls) == 1
    assert len(collector.lock.release.mock_calls) == 1
    assert collector.counters[count_name]["value"] == count_value
    assert collector.counters[count_name]["metadata"] == count_metadata
    assert collector.counters[count_name]["timestamp"] == NOW

@mock.patch('kadabra.client.get_now', return_value=NOW)
def test_collector_add_count_existing(mock_get_now):
    count_name = "name"
    count_value = 1.0
    timestamp_format = "timestamp_format"

    collector = kadabra.client.Collector(timestamp_format)
    collector.lock = MockLock()

    collector.add_count(count_name, count_value)
    collector.add_count(count_name, count_value)

    assert len(collector.lock.acquire.mock_calls) == 2
    assert len(collector.lock.release.mock_calls) == 2
    assert collector.counters[count_name]["value"] == count_value + count_value
    assert collector.counters[count_name]["metadata"] == {}
    assert collector.counters[count_name]["timestamp"] == NOW

@mock.patch('kadabra.client.get_now', return_value=NOW)
def test_collector_add_count_existing_replace_metadata(mock_get_now):
    count_name = "name"
    count_value = 1.0
    count_metadata = {"name": "value"}
    timestamp_format = "timestamp_format"

    collector = kadabra.client.Collector(timestamp_format)
    collector.lock = MockLock()

    collector.add_count(count_name, count_value, metadata=count_metadata)
    assert collector.counters[count_name]["metadata"] == count_metadata
    count_metadata_replacement = {"newName": "newValue"}
    collector.add_count(count_name, count_value,
            metadata=count_metadata_replacement)

    assert len(collector.lock.acquire.mock_calls) == 2
    assert len(collector.lock.release.mock_calls) == 2
    assert collector.counters[count_name]["value"] == count_value + count_value
    assert collector.counters[count_name]["metadata"] ==\
        count_metadata_replacement
    assert collector.counters[count_name]["timestamp"] == NOW

@mock.patch('kadabra.client.get_now', return_value=NOW)
def test_collector_add_count_existing_replace_timestamp(mock_get_now):
    count_name = "name"
    count_value = 1.0
    count_metadata = {"name": "value"}
    timestamp_format = "timestamp_format"
    timestamp = NOW + datetime.timedelta(seconds=30)

    collector = kadabra.client.Collector(timestamp_format)
    collector.lock = MockLock()

    collector.add_count(count_name, count_value)
    collector.add_count(count_name, count_value, timestamp=timestamp,
            replace_timestamp=True)

    assert len(collector.lock.acquire.mock_calls) == 2
    assert len(collector.lock.release.mock_calls) == 2
    assert collector.counters[count_name]["value"] == count_value + count_value
    assert collector.counters[count_name]["metadata"] == {}
    assert collector.counters[count_name]["timestamp"] == timestamp

@mock.patch('kadabra.client.get_now', return_value=NOW)
def test_collector_set_timer(mock_get_now):
    timer_name = "name"
    timer_value = datetime.timedelta(seconds=10)
    unit = kadabra.Units.MILLISECONDS
    timestamp_format = "timestamp_format"

    collector = kadabra.client.Collector(timestamp_format)
    collector.lock = MockLock()

    collector.set_timer(timer_name, timer_value, unit)

    assert len(collector.lock.acquire.mock_calls) == 1
    assert len(collector.lock.release.mock_calls) == 1
    assert collector.timers[timer_name]["value"] == timer_value
    assert collector.timers[timer_name]["unit"] == unit
    assert collector.timers[timer_name]["metadata"] == {}
    assert collector.timers[timer_name]["timestamp"] == NOW

def test_collector_set_timer_timestamp():
    timer_name = "name"
    timer_value = datetime.timedelta(seconds=10)
    unit = kadabra.Units.MILLISECONDS
    timestamp_format = "timestamp_format"
    timestamp = NOW + datetime.timedelta(seconds=30)

    collector = kadabra.client.Collector(timestamp_format)
    collector.lock = MockLock()

    collector.set_timer(timer_name, timer_value, unit, timestamp=timestamp)

    assert len(collector.lock.acquire.mock_calls) == 1
    assert len(collector.lock.release.mock_calls) == 1
    assert collector.timers[timer_name]["value"] == timer_value
    assert collector.timers[timer_name]["unit"] == unit
    assert collector.timers[timer_name]["metadata"] == {}
    assert collector.timers[timer_name]["timestamp"] == timestamp

@mock.patch('kadabra.client.get_now', return_value=NOW)
def test_collector_set_timer_metadata(mock_get_now):
    timer_name = "name"
    timer_value = datetime.timedelta(seconds=10)
    unit = kadabra.Units.MILLISECONDS
    timestamp_format = "timestamp_format"
    metadata = {"name": "value"}

    collector = kadabra.client.Collector(timestamp_format)
    collector.lock = MockLock()

    collector.set_timer(timer_name, timer_value, unit, metadata=metadata)

    assert len(collector.lock.acquire.mock_calls) == 1
    assert len(collector.lock.release.mock_calls) == 1
    assert collector.timers[timer_name]["value"] == timer_value
    assert collector.timers[timer_name]["unit"] == unit
    assert collector.timers[timer_name]["metadata"] == metadata
    assert collector.timers[timer_name]["timestamp"] == NOW

def test_collector_set_timer_existing():
    timer_name = "name"
    timer_value = datetime.timedelta(seconds=10)
    unit = kadabra.Units.MILLISECONDS
    timestamp_format = "timestamp_format"
    timestamp = NOW + datetime.timedelta(seconds=30)

    collector = kadabra.client.Collector(timestamp_format)
    collector.lock = MockLock()

    collector.set_timer(timer_name, timer_value, unit)
    unit_replacement = kadabra.Units.SECONDS
    timer_value_replacement = datetime.timedelta(seconds=20)
    collector.set_timer(timer_name, timer_value_replacement, unit_replacement,
            timestamp=timestamp)

    assert len(collector.lock.acquire.mock_calls) == 2
    assert len(collector.lock.release.mock_calls) == 2
    assert collector.timers[timer_name]["value"] == timer_value_replacement
    assert collector.timers[timer_name]["unit"] == unit_replacement
    assert collector.timers[timer_name]["metadata"] == {}
    assert collector.timers[timer_name]["timestamp"] == timestamp

def test_collector_set_timer_existing_replace_metadata():
    timer_name = "name"
    timer_value = datetime.timedelta(seconds=10)
    unit = kadabra.Units.MILLISECONDS
    metadata = {"name": "value"}
    timestamp_format = "timestamp_format"
    timestamp = NOW + datetime.timedelta(seconds=30)

    collector = kadabra.client.Collector(timestamp_format)
    collector.lock = MockLock()

    collector.set_timer(timer_name, timer_value, unit, metadata=metadata)
    unit_replacement = kadabra.Units.SECONDS
    timer_value_replacement = datetime.timedelta(seconds=20)
    metadata_replacement = {"newName": "newValue"}
    collector.set_timer(timer_name, timer_value_replacement, unit_replacement,
            timestamp=timestamp, metadata=metadata_replacement)

    assert len(collector.lock.acquire.mock_calls) == 2
    assert len(collector.lock.release.mock_calls) == 2
    assert collector.timers[timer_name]["value"] == timer_value_replacement
    assert collector.timers[timer_name]["unit"] == unit_replacement
    assert collector.timers[timer_name]["metadata"] == metadata_replacement
    assert collector.timers[timer_name]["timestamp"] == timestamp

# TODO: Test close() with mocked counters, timers, dimensions.
@mock.patch('kadabra.client.Dimension')
@mock.patch('kadabra.client.Counter')
@mock.patch('kadabra.client.Timer')
@mock.patch('kadabra.client.Metrics')
def test_collector_close(mock_metrics, mock_timer, mock_counter,
        mock_dimension):
    dimensions_expected = ['dimensionOne', 'dimensionTwo', 'dimensionThree']
    counters_expected = ['counterOne', 'counterTwo', 'counterThree']
    timers_expected = ['timerOne', 'timerTwo', 'timerThree']
    metrics_expected = "metrics"

    mock_dimension.side_effect = dimensions_expected
    mock_counter.side_effect = counters_expected
    mock_timer.side_effect = timers_expected
    mock_metrics.return_value = metrics_expected

    timestamp_format = "%Y-%m-%dT%H:%M:%S.%fZ"

    collector = kadabra.client.Collector(timestamp_format)
    collector.lock = MockLock()

    dimensions = [
        {
            "name": "dimNameOne",
            "value": "dimValueOne"
        },
        {
            "name": "dimNameTwo",
            "value": "dimValueTwo"
        },
        {
            "name": "dimNameThree",
            "value": "dimValueThree"
        }
    ]
    counters = [
        {
            "name": "countOne",
            "value": 1.0,
            "timestamp": NOW,
            "metadata": {"nameCountOne": "valueCountOne"}
        },
        {
            "name": "countTwo",
            "value": 1.0,
            "timestamp": NOW + datetime.timedelta(seconds=5),
            "metadata": {"nameCountTwo": "valueCountTwo"}
        },
        {
            "name": "countThree",
            "value": 1.0,
            "timestamp": NOW + datetime.timedelta(seconds=10),
            "metadata": {"nameCountThree": "valueCountThree"}
        }
    ]
    timers = [
        {
            "name": "timerOne",
            "value": datetime.timedelta(seconds=10),
            "timestamp": NOW,
            "unit": kadabra.Units.MILLISECONDS,
            "metadata": {"nameTimerOne": "valueTimerOne"}
        },
        {
            "name": "timerTwo",
            "value": datetime.timedelta(seconds=20),
            "timestamp": NOW + datetime.timedelta(seconds=10),
            "unit": kadabra.Units.MILLISECONDS,
            "metadata": {"nameTimerTwo": "valueTimerTwo"}
        },
        {
            "name": "timerThree",
            "value": datetime.timedelta(seconds=30),
            "timestamp": NOW + datetime.timedelta(seconds=20),
            "unit": kadabra.Units.MILLISECONDS,
            "metadata": {"nameTimerThree": "valueTimerThree"}
        }
    ]

    for d in dimensions:
        collector.set_dimension(**d)
    for c in counters:
        collector.add_count(**c)
    for t in timers:
        collector.set_timer(**t)
    metrics = collector.close()

    assert metrics == metrics_expected

    assert len(collector.lock.acquire.mock_calls) ==\
            len(dimensions) + len(counters) + len(timers) + 1
    assert len(collector.lock.release.mock_calls) ==\
            len(dimensions) + len(counters) + len(timers) + 1

    mock_dimension.assert_has_calls([call(d["name"], d["value"])\
            for d in dimensions], any_order=True)
    mock_counter.assert_has_calls([call(c["name"], c["timestamp"],
        c["metadata"], c["value"]) for c in counters], any_order=True)
    mock_timer.assert_has_calls([call(t["name"], t["timestamp"],
        t["metadata"], t["value"], t["unit"]) for t in timers],
        any_order=True)

    mock_metrics.assert_called_with(dimensions_expected, counters_expected,
            timers_expected, timestamp_format)



from .utils import get_now, timedelta_total_seconds

import datetime

class Dimension(object):
    """Dimensions are used for grouping sets of metrics by shared components.
    They are key-value string pairs which are meant to be indexed in the
    metrics storage for ease of querying metrics.

    :type name: string
    :param name: The name of the dimension.

    :type value: string
    :param value: The value of the dimension.
    """
    def __init__(self, name, value):
        self.name = name
        self.value = value

    def serialize(self):
        """Serializes this dimension to a dictionary.

        :rtype: dict
        :returns: The dimension as a dictionary.
        """
        return {
            "name": self.name,
            "value": self.value
        }

    @staticmethod
    def deserialize(value):
        """Deserializes a dictionary into a :class:`~kadabra.Dimension`
        instance.

        :type value: dict
        :param value: The dictionary to deserialize into a
                      :class:`~kadabra.Dimension` instance.

        :rtype: ~kadabra.Dimension
        :returns: A dimension that the dictionary represents.
        """
        return Dimension(value["name"], value["value"])

class Metric(object):
    """Base class for all metric types. Should not be instantiated directly.

    :type name: string
    :param name: The name of the metric.

    :type timestamp: ~datetime.datetime
    :param timestamp: The timestamp of the metric.

    :type metadata: dict
    :param metadata: Metadata associated with this metric, in the form of
                     string-string key-value pairs. This metadata is meant to
                     be stored as non-indexed fields in the metrics storage.
    """
    def __init__(self, name, timestamp, metadata):
        self.name = name
        self.timestamp = timestamp
        self.metadata = metadata

class Counter(Metric):
    """A counter metric, which consists of a name and a floating-point
    value.

    :type name: string
    :param name: The name of the metric.

    :type timestamp: ~datetime.datetime
    :param timestamp: The timestamp of the metric.

    :type metadata: dict
    :param metadata: Metadata associated with this metric, in the form of
                     string-string key-value pairs. This metadata is meant to
                     be stored as non-indexed fields in the metrics storage.

    :type value: float
    :param value: The floating-point value of this counter.
    """
    def __init__(self, name, timestamp, metadata, value):
        super(Counter, self).__init__(name, timestamp, metadata)
        self.value = value

    def serialize(self, timestamp_format):
        """Serializes this counter to a dictionary.

        :type timestamp_format: string
        :param timestamp_format: The format string for this counter's timestamp.

        :rtype: dict
        :returns: The counter as a dictionary.
        """
        return {
            "name": self.name,
            "metadata": self.metadata,
            "timestamp": self.timestamp.strftime(timestamp_format),
            "value": float(self.value)
        }

    @staticmethod
    def deserialize(value, timestamp_format):
        """Deserializes a dictionary into a :class:`~kadabra.Counter`
        instance.

        :type value: dict
        :param value: The dictionary to deserialize into a
                      :class:`~kadabra.Counter` instance.

        :rtype: ~kadabra.Counter
        :returns: A counter that the dictionary represents.
        """
        return Counter(value["name"],
                datetime.datetime.strptime(value["timestamp"],
                    timestamp_format),
                value["metadata"],
                value["value"])

class Timer(Metric):
    """A timer metric representing an elapsed period of time, identified by a
    :class:`datetime.timedelta` and a :class:`~kadabra.Unit`.

    :type name: string
    :param name: The name of the timer.

    :type timestamp: ~datetime.datetime
    :param timestamp: The timestamp of the timer.

    :type metadata: dict
    :param metadata: The metadata associated with the timer.

    :type value: ~datetime.timedelta
    :param value: The value of the timer.

    :type unit: kadabra.Unit
    :param unit: The unit of the timer value.
    """
    def __init__(self, name, timestamp, metadata, value, unit):
        self.value = value
        self.unit = unit
        super(Timer, self).__init__(name, timestamp, metadata)

    def serialize(self, timestamp_format):
        """Serializes this timer to a dictionary.

        :type timestamp_format: string
        :param timestamp_format: The format string for this timer's timestamp.

        :rtype: dict
        :returns: The timer as a dictionary.
        """
        return {
            'name': self.name,
            'unit': self.unit.serialize(),
            'metadata': self.metadata,
            'timestamp': self.timestamp.strftime(timestamp_format),
            'value': timedelta_total_seconds(self.value) *\
                    self.unit.seconds_offset
        }

    @staticmethod
    def deserialize(value, timestamp_format):
        """Deserializes a dictionary into a :class:`~kadabra.Timer` instance.

        :type value: dict
        :param value: The dictionary to deserialize into a
                      :class:`~kadabra.Timer` instance.

        :rtype: ~kadabra.Timer
        :returns: A timer that the dictionary represents.
        """
        unit = Unit.deserialize(value["unit"])
        seconds = value["value"] / unit.seconds_offset
        timer_value = datetime.timedelta(seconds=seconds)
        return Timer(value["name"],
                datetime.datetime.strptime(value["timestamp"],
                    timestamp_format),
                value["metadata"],
                timer_value,
                unit)

class Unit(object):
    """A unit, representing an offset from seconds. This is used by by
    :class:`kadabra.Timer`\s for unambiguous reporting of the timer's value.

    :type name: string
    :param name: The name of the unit.

    :type seconds_offset: integer
    :param seconds_offset: The offset of the unit relative to seconds.
    """
    def __init__(self, name, seconds_offset):
        self.name = name
        self.seconds_offset = seconds_offset

    def serialize(self):
        """Serializes this unit to a dictionary.

        :rtype: dict
        :returns: The unit as a dictionary.
        """
        return {
            "name": self.name,
            "seconds_offset": self.seconds_offset
        }

    @staticmethod
    def deserialize(value):
        """Deserializes a dictionary into a :class:`~kadabra.Unit` instance.

        :type value: dict
        :param value: The dictionary to deserialize into a
                      :class:`~kadabra.Unit` instance.

        :rtype: ~kadabra.Unit
        :returns: A unit that the dictionary represents.
        """
        return Unit(value["name"], value["seconds_offset"])

class Units(object):
    """Container for commonly used units."""
    #: Unit representing seconds.
    SECONDS = Unit("seconds", 1.0)

    #: Unit representing milliseconds.
    MILLISECONDS = Unit("milliseconds", 1000.0)

class Metrics(object):
    """This class encapsulates metrics which can be transported over a channel,
    and received by the agent. It should only ever be initialized (e.g.
    instances are meant to be immutable). This guarantees correct behavior with
    respect to the client (which transports the metrics) and the agent (which
    receives and publishes the metrics).

    :type dimensions: list
    :param dimensions: :class:`~kadabra.Dimension`\s for this set of metrics.

    :type counters: list
    :param counters: :class:`~kadabra.Counter`\s for this set of metrics.

    :type timers: list
    :param timers: :class:`~kadabra.Timer`\s for this set of metrics.

    :type timestamp_format: string
    :param timestamp_format: The format string for timestamps.

    :type serialized_at: string
    :param serialized_at: The timestamp string for when the metrics were
                          serialized, if they were previously serialized.
    """
    def __init__(self, dimensions, counters, timers,
            timestamp_format="%Y-%m-%dT%H:%M:%S.%fZ",
            serialized_at=None):
        self.dimensions = dimensions
        self.counters = counters
        self.timers = timers

        self.timestamp_format = timestamp_format
        self.serialized_at = serialized_at

    def serialize(self):
        """Serializes this set of metrics into a dictionary.

        :rtype: dict
        :returns: The metrics as a dictionary.
        """
        return {
            "dimensions": [d.serialize() for d in self.dimensions],
            "counters": [c.serialize(self.timestamp_format)\
                    for c in self.counters],
            "timers": [t.serialize(self.timestamp_format)\
                    for t in self.timers],
            "timestamp_format": self.timestamp_format,
            "serialized_at":  get_now().strftime(\
                    self.timestamp_format) if self.serialized_at is None\
                    else self.serialized_at
        }

    @staticmethod
    def deserialize(value):
        """Deserializes a dictionary into a :class:`~kadabra.Metrics`
        instance.

        :type value: dict
        :param value: The dictionary to deserialize into a
                      :class:`~kadabra.Metrics` instance.

        :rtype: ~kadabra.Metrics
        :returns: A metrics that the dictionary represents.
        """
        timestamp_format = value["timestamp_format"]
        serialized_at = value["serialized_at"]\
                if "serialized_at" in value else None
        return Metrics(\
                [Dimension.deserialize(d) for d in value["dimensions"]],
                [Counter.deserialize(c, timestamp_format)\
                        for c in value["counters"]],
                [Timer.deserialize(t, timestamp_format)\
                        for t in value["timers"]],
                timestamp_format,
                serialized_at)




from .utils import timedelta_total_seconds

import sys, json, datetime

class DebugPublisher(object):
    """Publish metrics to a logger using the given logger name. Useful for
    debugging.

    :type logger_name: string
    :param logger_name: The name of the logger to use.
    """

    #: Default arguments for this publisher. These will be used by the
    #: agent to initialize this publisher if custom configuration values are
    #: not provided.
    DEFAULT_ARGS = {"logger_name": "kadabra.publisher"}

    def __init__(self, logger_name):
        import logging
        self.logger = logging.getLogger(logger_name)

    def publish(self, metrics):
        """Publish the metrics by logging them (in serialized JSON format) to
        the publisher's logger at the INFO level.

        :type metrics: ~kadabra.Metrics
        :param metrics: The metrics to publish.
        """
        self.logger.info(metrics.serialize())

class InfluxDBPublisher(object):
    """Publish metrics by persisting them into an InfluxDB database. Series
    will be created for each metric. Each metric name becomes a measurement
    and dimensions become the tag set. A single field will be created called
    'value' which contains the value of the counter or timer. Timers will have
    an additional field called 'unit' which contains the name of the unit. Any
    metadata will become additional fields, although note that 'value' is a
    reserved name that will be overwritten for both metric types, and 'unit'
    will be overwritten for timers. For more information about InfluxDB see
    the `docs <https://docs.influxdata.com/influxdb>`.

    :type host: string
    :param host: The hostname of the InfluxDB database.

    :type port: int
    :param port: The port of the InfluxDB database.

    :type database: string
    :param database: The name of the database to use for publishing metrics
                     with this publisher. Note that this database must exist
                     prior to publishing metrics with this publisher - make
                     sure you set it up beforehand!

    :type timeout: int
    :param timeout: The timeout to wait for when calling the InfluxDB database
                    before failing.
    """

    #: Default arguments for this publisher. These will be used by the
    #: agent to initialize this publisher if custom configuration values are
    #: not provided.
    DEFAULT_ARGS = {"host": "localhost", "port": 8086,\
            "database": "kadabra", "timeout": 5}

    def __init__(self, host, port, database, timeout):
        from influxdb import InfluxDBClient
        self.client = InfluxDBClient(host=host, port=port, database=database,\
                timeout=timeout)

    def publish(self, metrics):
        """Publish the metrics by writing them to InfluxDB.

        :type metrics: ~kadabra.Metrics
        :param metrics: The metrics to publish.
        """
        data = []
        tags = dict([(d.name, d.value) for d in metrics.dimensions])

        for timer in metrics.timers:
            fields = dict([(k, v) for k,v in timer.metadata.items()])
            fields["value"] = timedelta_total_seconds(timer.value) *\
                    timer.unit.seconds_offset
            fields["unit"] = timer.unit.name
            datum = {
                "measurement": timer.name,
                "tags": tags,
                "time": datetime.datetime.strftime(timer.timestamp,
                    metrics.timestamp_format),
                "fields": fields
            }
            data.append(datum)

        for counter in metrics.counters:
            fields = dict([(k, v) for k,v in counter.metadata.items()])
            fields["value"] = counter.value
            datum = {
                "measurement": counter.name,
                "tags": tags,
                "time": datetime.datetime.strftime(counter.timestamp,
                    metrics.timestamp_format),
                "fields": fields
            }
            data.append(datum)

        if len(data) > 0:
            self.client.write_points(data)



import datetime

def get_now():
    return datetime.datetime.utcnow()

def get_datetime_from_timestamp_string(timestamp_string, timestamp_format):
    return datetime.datetime.strptime(timestamp_string, timestamp_format)

def timedelta_total_seconds(td):
    return (td.microseconds + (td.seconds + td.days * 24 * 3600) * 10**6) /\
            10.0**6



"""
    tests.metrics
    ~~~~~~~~~~~
    Definitions of basic metric classes and serialization/deserialization.

    :copyright: (c) 2016 by Alex Landau.
    :license: BSD, see LICENSE for more details.
"""

import datetime
import kadabra

from mock import MagicMock, mock

def test_dimension():
    name = "dimensionName"
    value = "dimensionValue"
    expected_serialized = {
        "name": name,
        "value": value
    }

    dimension = kadabra.Dimension(name, value)

    assert dimension.name == name
    assert dimension.value == value
    assert dimension.serialize() == expected_serialized

    deserialized = kadabra.Dimension.deserialize(expected_serialized)
    assert deserialized.name == name
    assert deserialized.value == value

def test_counter():
    name = "myCounter"
    timestamp = datetime.datetime.utcnow()
    timestamp_format = "%Y-%m-%dT%H:%M:%S.%fZ"
    metadata = {"mdName": "mdValue"}
    value = 15.0

    expected_serialized = {
        "name": name,
        "timestamp": timestamp.strftime(timestamp_format),
        "metadata": metadata,
        "value": value
    }

    counter = kadabra.Counter(name, timestamp, metadata, value)

    assert counter.name == name
    assert counter.timestamp == timestamp
    assert counter.metadata == metadata
    assert counter.value == value
    assert counter.serialize(timestamp_format) == expected_serialized

    deserialized = kadabra.Counter.deserialize(expected_serialized,
            timestamp_format)
    assert deserialized.name == name
    assert deserialized.timestamp == timestamp
    assert deserialized.metadata == metadata
    assert deserialized.value == value

def test_unit():
    name = "unitName"
    seconds_offset = 25.0
    expected_serialized = {
        "name": name,
        "seconds_offset": seconds_offset
    }

    unit = kadabra.Unit(name, seconds_offset)

    assert unit.name == name
    assert unit.seconds_offset == seconds_offset
    assert unit.serialize() == expected_serialized

    deserialized = kadabra.Unit.deserialize(expected_serialized)
    assert deserialized.name == name
    assert deserialized.seconds_offset == seconds_offset

@mock.patch('kadabra.Unit.deserialize')
def test_timer(mock_deserialize):
    name = "timerName"
    timestamp = datetime.datetime.utcnow()
    timestamp_format = "%Y-%m-%dT%H:%M:%S.%fZ"
    metadata = {"mdName": "mdValue"}
    value = datetime.timedelta(seconds=5)
    unit_seconds_offset = 25.0
    unit = kadabra.Unit("unitName", unit_seconds_offset)

    unit.serialize = MagicMock(return_value="unit")
    mock_deserialize.return_value = unit

    expected_serialized = {
        "name": name,
        "metadata": metadata,
        "timestamp": timestamp.strftime(timestamp_format),
        "unit": "unit",
        "value": kadabra.utils.timedelta_total_seconds(value) *\
                unit_seconds_offset
    }

    timer = kadabra.Timer(name, timestamp, metadata, value, unit)

    assert timer.name == name
    assert timer.timestamp == timestamp
    assert timer.metadata == metadata
    assert timer.value == value
    assert timer.unit == unit
    assert timer.serialize(timestamp_format) == expected_serialized

    deserialized = kadabra.Timer.deserialize(expected_serialized,
            timestamp_format)

    mock_deserialize.assert_called_with("unit")

    assert deserialized.name == name
    assert deserialized.timestamp == timestamp
    assert deserialized.metadata == metadata
    assert deserialized.value == value
    assert deserialized.unit == unit

def test_metrics_serialize():
    dimension_one = kadabra.Dimension("name", "value")
    dimension_two = kadabra.Dimension("name", "value")
    dimension_three = kadabra.Dimension("name", "value")

    dimension_one.serialize = MagicMock(return_value='dimensionOne')
    dimension_two.serialize = MagicMock(return_value='dimensionTwo')
    dimension_three.serialize = MagicMock(return_value='dimensionThree')

    counter_one = kadabra.Counter("name", "time", {}, 1.0)
    counter_two = kadabra.Counter("name", "time", {}, 1.0)
    counter_three = kadabra.Counter("name", "time", {}, 1.0)

    counter_one.serialize = MagicMock(return_value='counterOne')
    counter_two.serialize = MagicMock(return_value='counterTwo')
    counter_three.serialize = MagicMock(return_value='counterThree')

    timer_one = kadabra.Timer("name", "time", {}, "value", "unit")
    timer_two = kadabra.Timer("name", "time", {}, "value", "unit")
    timer_three = kadabra.Timer("name", "time", {}, "value", "unit")

    timer_one.serialize = MagicMock(return_value='timerOne')
    timer_two.serialize = MagicMock(return_value='timerTwo')
    timer_three.serialize = MagicMock(return_value='timerThree')

    dimensions = [dimension_one, dimension_two, dimension_three]
    counters = [counter_one, counter_two, counter_three]
    timers = [timer_one, timer_two, timer_three]

    timestamp_format = "%Y-%m-%dT%H:%M:%S.%fZ"
    serialized_at = "now"

    expected_serialized = {
        "dimensions": ['dimensionOne', 'dimensionTwo', 'dimensionThree'],
        "counters": ['counterOne', 'counterTwo', 'counterThree'],
        "timers": ['timerOne', 'timerTwo', 'timerThree'],
        "timestamp_format": timestamp_format,
        "serialized_at": serialized_at
    }

    metrics = kadabra.Metrics(dimensions, counters, timers, timestamp_format)
    assert metrics.dimensions == dimensions
    assert metrics.counters == counters
    assert metrics.timers == timers
    assert metrics.timestamp_format == timestamp_format
    assert metrics.serialized_at == None

    serialized = metrics.serialize()
    assert serialized["dimensions"] == expected_serialized["dimensions"]
    assert serialized["counters"] == expected_serialized["counters"]
    assert serialized["timers"] == expected_serialized["timers"]
    assert serialized["timestamp_format"] ==\
            expected_serialized["timestamp_format"]

    # Can't figure out how to mock utcnow() cleanly, so just make sure it
    # parses to a valid date w.r.t. the timestamp format.
    datetime.datetime.strptime(serialized["serialized_at"], timestamp_format)

    # Make sure if we specify a serialized_at we get it back
    metrics = kadabra.Metrics(dimensions, counters, timers, timestamp_format,
            "test")
    assert metrics.serialize()["serialized_at"] == "test"

# Note: multiple mock patches apply to the function args in reverse order.
@mock.patch("kadabra.Dimension.deserialize")
@mock.patch("kadabra.Counter.deserialize")
@mock.patch("kadabra.Timer.deserialize")
def test_metrics_deserialize(mock_timer_deserialize,
        mock_counter_deserialize, mock_dimension_deserialize):
    dimensions = ['dimensionOne', 'dimensionTwo', 'dimensionThree']
    counters = ['counterOne', 'counterTwo', 'counterThree']
    timers = ['timerOne', 'timerTwo', 'timerThree']

    mock_dimension_deserialize.side_effect = dimensions
    mock_counter_deserialize.side_effect = counters
    mock_timer_deserialize.side_effect = timers

    timestamp_format = "%Y-%m-%dT%H:%M:%S.%fZ"
    serialized_at = datetime.datetime.utcnow().strftime(timestamp_format)

    metrics_serialized = {
        "dimensions": dimensions,
        "counters": counters,
        "timers": timers,
        "timestamp_format": timestamp_format,
        "serialized_at": serialized_at
    }

    metrics = kadabra.Metrics.deserialize(metrics_serialized)
    assert metrics.dimensions == dimensions
    assert metrics.counters == counters
    assert metrics.timers == timers
    assert metrics.timestamp_format == timestamp_format
    assert metrics.serialized_at == serialized_at

@mock.patch("kadabra.Dimension.deserialize")
@mock.patch("kadabra.Counter.deserialize")
@mock.patch("kadabra.Timer.deserialize")
def test_metrics_deserialize_no_serialized_at(mock_timer_deserialize,
        mock_counter_deserialize, mock_dimension_deserialize):
    dimensions = ['dimensionOne', 'dimensionTwo', 'dimensionThree']
    counters = ['counterOne', 'counterTwo', 'counterThree']
    timers = ['timerOne', 'timerTwo', 'timerThree']

    mock_dimension_deserialize.side_effect = dimensions
    mock_counter_deserialize.side_effect = counters
    mock_timer_deserialize.side_effect = timers

    timestamp_format = "%Y-%m-%dT%H:%M:%S.%fZ"

    metrics_serialized = {
        "dimensions": dimensions,
        "counters": counters,
        "timers": timers,
        "timestamp_format": timestamp_format
    }

    metrics = kadabra.Metrics.deserialize(metrics_serialized)
    assert metrics.dimensions == dimensions
    assert metrics.counters == counters
    assert metrics.timers == timers
    assert metrics.timestamp_format == timestamp_format
    assert metrics.serialized_at == None



"""
Kadabra
-------

Kadabra makes it easy to publish metrics from your Python application. It
provides a formal definition for what metrics should look like, and channels
to efficiently publish metrics from your python application to metrics storage.  
"""
import ast, re

from setuptools import setup

_version_re = re.compile(r'__version__\s+=\s+(.*)')

with open('kadabra/__init__.py', 'rb') as f:
    version = str(ast.literal_eval(_version_re.search(
        f.read().decode('utf-8')).group(1)))

setup(
    name='Kadabra',
    version=version,
    url='http://github.com/bal2ag/kadabra/',
    license='BSD',
    author='Alex Landau',
    author_email='toozlyllc@gmail.com',
    description='A simple interface for publishing application metrics',
    long_description=__doc__,
    packages=['kadabra'],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'influxdb>=3.0.0',
        'redis>=2.10',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)
    



import kadabra

from mock import MagicMock, call

def test_ctor():
    channel = MagicMock()
    publisher = MagicMock()
    logger = MagicMock()

    receiver_thread = kadabra.agent.ReceiverThread(channel, publisher, logger)

    assert receiver_thread.channel == channel
    assert receiver_thread.publisher == publisher
    assert receiver_thread.logger == logger

def test_run_once_no_metrics():
    channel = MagicMock()
    publisher = MagicMock()
    logger = MagicMock()

    channel.receive = MagicMock(return_value=None)
    publisher.publish = MagicMock()
    channel.complete = MagicMock()

    receiver_thread = kadabra.agent.ReceiverThread(channel, publisher, logger)
    receiver_thread._run_once()

    channel.receive.assert_called_with()
    publisher.publish.assert_has_calls([])
    channel.complete.assert_has_calls([])

def test_run_once_metrics():
    channel = MagicMock()
    publisher = MagicMock()
    logger = MagicMock()

    channel.receive = MagicMock()
    metrics = channel.receive.return_value
    metrics.serialize = MagicMock()
    publisher.publish = MagicMock()
    channel.complete = MagicMock()

    receiver_thread = kadabra.agent.ReceiverThread(channel, publisher, logger)
    receiver_thread._run_once()

    channel.receive.assert_called_with()
    metrics.serialize.assert_called_with()
    publisher.publish.assert_called_with(metrics)
    channel.complete.assert_called_with(metrics)

def test_run_once_exception():
    channel = MagicMock()
    publisher = MagicMock()
    logger = MagicMock()

    channel.receive = MagicMock()
    metrics = channel.receive.return_value
    metrics.serialize = MagicMock()
    publisher.publish = MagicMock()
    publisher.publish.side_effect = Exception()
    channel.complete = MagicMock()

    receiver_thread = kadabra.agent.ReceiverThread(channel, publisher, logger)
    receiver_thread._run_once()

    channel.receive.assert_called_with()
    metrics.serialize.assert_called_with()
    publisher.publish.assert_called_with(metrics)
    channel.complete.assert_has_calls([])

def test_stop():
    channel = MagicMock()
    publisher = MagicMock()
    logger = MagicMock()

    receiver_thread = kadabra.agent.ReceiverThread(channel, publisher, logger)
    assert receiver_thread.stopped == False
    receiver_thread.stop()
    assert receiver_thread.stopped == True

def test_check_stopped():
    channel = MagicMock()
    publisher = MagicMock()
    logger = MagicMock()

    receiver_thread = kadabra.agent.ReceiverThread(channel, publisher, logger)
    receiver_thread.stopped = False

    assert receiver_thread._check_stopped() == False
    receiver_thread.stopped = True
    assert receiver_thread._check_stopped() == True

def test_run():
    channel = MagicMock()
    publisher = MagicMock()
    logger = MagicMock()

    receiver_thread = kadabra.agent.ReceiverThread(channel, publisher, logger)
    receiver_thread._run_once = MagicMock()
    receiver_thread._check_stopped = MagicMock()
    receiver_thread._check_stopped.side_effect = [False, True]

    receiver_thread.run()

    assert receiver_thread._check_stopped.call_count == 2
    assert receiver_thread._run_once.call_count == 1



from .metrics import Metrics

import logging, json

class RedisChannel(object):
    """A channel for transporting metrics using Redis.

    :type host: string
    :param host: The host of the Redis server.

    :type port: int
    :param port: The port of the Redis server.

    :type db: int
    :param db: The database to use on the Redis server. This should be used
               exclusively for Kadabra to prevent collisions with keys that
               might be used by your application.

    :type logger: string
    :param logger: The name of the logger to use.
    """

    #: Default arguments for the Redis channel. These will be used by the
    #: client and agent to initialize this channel if custom configuration
    #: values are not provided.
    DEFAULT_ARGS = {
            "host": "localhost",
            "port": 6379,
            "db": 0,
            "logger": "kadabra.channel",
            "queue_key": "kadabra_queue",
            "inprogress_key": "kadabra_inprogress"
    }

    def __init__(self, host, port, db, logger, queue_key, inprogress_key):
        from redis import StrictRedis
        self.client = StrictRedis(host=host, port=port, db=db)
        self.logger = logging.getLogger(logger)
        self.queue_key = queue_key
        self.inprogress_key = inprogress_key

    def send(self, metrics):
        """Send metrics to a Redis list, which will act as queue for pending
        metrics to be received and published.

        :type metrics: ~kadabra.Metrics
        :param metrics: The metrics to be sent.
        """
        to_push = metrics.serialize()
        self.logger.debug("Sending %s" % to_push)
        self.client.lpush(self.queue_key, json.dumps(to_push))
        self.logger.debug("Successfully sent %s" % to_push)

    def receive(self):
        """Receive metrics from the queue so they can be published. Once
        received, the metrics will be moved into a temporary "in progress"
        queue until they have been acknowledged as published (by calling
        :meth:`~kadabra.channels.RedisChannel.complete`). This method will
        block until there are metrics available on the queue or after 10
        seconds.

        :rtype: ~kadabra.Metrics
        :returns: The metrics to be published, or None if there were no metrics
                  received after the timeout.
        """
        self.logger.debug("Receiving metrics")
        raw = self.client.brpoplpush(self.queue_key, self.inprogress_key,
                timeout=10)
        if raw:
            rv = json.loads(raw)
            self.logger.debug("Got metrics: %s" % rv)
            return Metrics.deserialize(rv)
        self.logger.debug("No metrics received")
        return None

    def complete(self, metrics):
        """Mark metrics as completed by removing them from the in-progress
        queue.

        :type metrics: ~kadabra.Metrics
        :param metrics: The metris to mark as complete.
        """
        to_complete = metrics.serialize()
        self.logger.debug("Marking %s as complete" % str(to_complete))
        rv = self.client.lrem(self.inprogress_key, 1, json.dumps(to_complete))
        if rv > 0:
            self.logger.debug("Successfully marked %s as complete" %\
                    str(to_complete))
        else:
            self.logger.debug("Failed to mark %s as complete" %\
                    str(to_complete))

    def in_progress(self, query_limit):
        """Return a list of the metrics that are in_progress.

        :type query_limit: int
        :param query_limit: The maximum number of items to get from the in
                            progress queue.

        :rtype: list
        :returns: A list of :class:`Metric`\s that are in progress.
        """
        in_progress = self.client.lrange(self.inprogress_key, 0,\
                query_limit - 1)
        self.logger.debug("Found %s in progress metrics" % len(in_progress))
        return [Metrics.deserialize(json.loads(m))\
                for m in in_progress]

